"""Training related classes and functions."""

import collections
import contextlib
import itertools
import time

import tensorflow as tf

from opennmt.inputters import text_inputter
from opennmt.optimizers import utils as optimizer_util
from opennmt.utils import misc


class Trainer:
    """Base class for model trainer, implementing single-GPU training."""

    def __init__(self, model, optimizer, checkpoint=None):
        """Initializes the trainer.

        Args:
          model: A :class:`opennmt.models.Model` instance to train.
          optimizer: A ``tf.keras.optimizers.Optimizer`` instance.
          checkpoint: A :class:`opennmt.utils.checkpoint.Checkpoint` instance. If
            not set, no checkpoints will be saved.
        """
        self._checkpoint = checkpoint
        self._model = model
        if checkpoint is not None:
            self._summary_writer = tf.summary.create_file_writer(checkpoint.model_dir)
        else:
            self._summary_writer = tf.summary.create_noop_writer()
        self._training_stats = None
        self._gradient_accumulator = optimizer_util.GradientAccumulator()
        self._mixed_precision = misc.mixed_precision_enabled()

        if optimizer is None:
            raise ValueError("No optimizer is defined")
        if self._mixed_precision and not isinstance(
            optimizer, tf.keras.mixed_precision.LossScaleOptimizer
        ):
            optimizer = tf.keras.mixed_precision.LossScaleOptimizer(optimizer)
        self._optimizer = optimizer

    @property
    def is_master(self):
        """Master replica."""
        return True

    @property
    def num_replicas(self):
        """Number of synchronous training replicas."""
        return 1

    def __call__(
        self,
        dataset,
        max_step=None,
        accum_steps=1,
        report_steps=100,
        save_steps=5000,
        evaluator=None,
        eval_steps=5000,
        moving_average_decay=None,
    ):
        """Runs the training.

        Args:
          dataset: A ``tf.data.Dataset`` or a function taking a ``tf.distribute.InputContext``
            instance and returning a ``tf.data.Dataset``.
          max_step: The final training step.
          accum_steps: The number of gradient accumulation steps.
          report_steps: Report status every this many steps.
          save_steps: Save a checkpoint every this many steps.
          evaluator: A :class:`opennmt.evaluation.Evaluator` instance to call for
            evaluation.
          eval_steps: Evaluate every this many steps.
          moving_average_decay: If set, maintain an exponential moving average of the model
            variables using this decay value (usually close to 1, e.g. 0.9999). See
            https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage.

        Returns:
          A dictionary with various training statistics.
        """
        if max_step is not None and self._optimizer.iterations.numpy() >= max_step:
            raise RuntimeError(
                "The training already reached max_step (%d). If you "
                "want to continue the training, you should increase the "
                "max_step value in the training parameters." % max_step
            )
        if evaluator is not None and evaluator.should_stop():
            raise RuntimeError(
                "The early stopping conditions are already met. If you "
                "want to continue the training, you should update your "
                "early stopping parameters."
            )

        self._gradient_accumulator.reset()

        with self._summary_writer.as_default():
            self._training_stats = TrainingStats(
                self._model, self._optimizer, reduce_fn=self._all_reduce_sum
            )
            iterations = self._optimizer.iterations
            tf.summary.experimental.set_step(iterations)

            step = None
            moving_average = None
            for i, loss in enumerate(
                self._steps(dataset, accum_steps=accum_steps, report_steps=report_steps)
            ):
                if i == 0:
                    self._log_model_info()

                if moving_average_decay is not None and self.is_master:
                    if moving_average is None:
                        moving_average = MovingAverage(
                            self._model.trainable_variables,
                            iterations,
                            decay=moving_average_decay,
                        )
                    self._update_moving_average(moving_average)

                step = iterations.numpy()
                reset_throughput = False
                self._training_stats.update_on_step(step, loss)
                if step % report_steps == 0:
                    self._training_stats.log(self.is_master)
                    reset_throughput = True
                if step == 1 or (save_steps is not None and step % save_steps == 0):
                    self._save_checkpoint(step, moving_average=moving_average)
                    reset_throughput = True
                if eval_steps is not None and step % eval_steps == 0:
                    early_stop = self._evaluate(
                        evaluator, step, moving_average=moving_average
                    )
                    reset_throughput = True
                    if early_stop:
                        tf.get_logger().warning(
                            "Early stopping conditions are met. Exiting."
                        )
                        break
                if step == max_step:
                    break
                if reset_throughput:
                    self._training_stats.reset_throughput()

            if step is None:
                raise RuntimeError(
                    "No training steps were executed. This usually means the "
                    "training file is empty or all examples were filtered out. "
                    "For the latter, verify that the maximum_*_length values are "
                    "consistent with your data."
                )

            self._training_stats.log_final(self.is_master)
            summary = self._training_stats.get_global_summary()
            self._save_checkpoint(step, moving_average=moving_average)
            self._evaluate(evaluator, step, moving_average=moving_average)
            return summary

    def _save_checkpoint(self, step, moving_average=None):
        """Saves a checkpoint for step."""
        if (
            not self.is_master
            or self._checkpoint is None
            or step == self._checkpoint.last_saved_step
        ):
            return
        shadow_variables = (
            moving_average.shadow_variables()
            if moving_average is not None
            else contextlib.suppress()
        )
        with shadow_variables:
            self._checkpoint.save(step)

    def _evaluate(self, evaluator, step, moving_average=None):
        """Runs evaluation for step. Returns ``True`` if early conditions are met."""
        if (
            not self.is_master
            or evaluator is None
            or step == evaluator.last_evaluated_step
        ):
            return False
        shadow_variables = (
            moving_average.shadow_variables()
            if moving_average is not None
            else contextlib.suppress()
        )
        with shadow_variables:
            evaluator(step)
            return evaluator.should_stop()

    def _finalize_dataset(self, dataset):
        """Returns the final dataset instance to be used for training.

        Args:
          dataset: A ``tf.data.Dataset`` or a function taking a ``tf.distribute.InputContext``
            instance and returning a ``tf.data.Dataset``.

        Returns:
          A ``tf.data.Dataset``.
        """
        if callable(dataset):
            dataset = dataset(tf.distribute.InputContext())
        return dataset

    def _steps(self, dataset, accum_steps=1, report_steps=None):
        """Returns a generator over training steps (i.e. parameters update).

        Args:
          dataset: The training dataset.
          accum_steps: Accumulate the gradients of this many steps/batches.
          report_steps: Report summary statistics every this many steps. This should
            typically be used in a ``tf.summary.record_if`` context.

        Returns:
          A generator that yields a loss value to report for this step.
        """
        dataset = self._finalize_dataset(dataset)
        iterator = iter(dataset)

        # We define 2 separate functions to support gradient accumulation:
        #  * forward: compute and accumulate the gradients
        #  * step: apply the gradients
        # When gradient accumulation is disabled, the forward function also applies the gradients.

        def _forward():
            # We get the next dataset element within the function for increased efficiency
            # and avoid dealing with tf.function input signatures.
            source, target = next(iterator)
            return self._forward(
                source,
                target,
                accum_steps=accum_steps,
                report_steps=report_steps,
            )

        def _step():
            return self._step()

        # Wrap forward and step with tf.function to run in graph mode.
        forward_fn = tf.function(_forward)
        step_fn = tf.function(_step) if accum_steps > 1 else lambda: None
        step_loss = 0

        for i in itertools.count():
            try:
                loss = forward_fn()
            except (
                StopIteration,
                tf.errors.OutOfRangeError,
            ):  # Dataset iterator exhausted.
                break

            if tf.math.is_nan(loss):
                raise RuntimeError("Model diverged with loss = NaN.")

            step_loss += float(loss)
            if (i + 1) % accum_steps == 0:
                step_fn()
                if i + 1 == accum_steps:
                    self._broadcast_variables()
                yield step_loss
                step_loss = 0

    def _log_model_info(self):
        """Logs some information about the model being trained."""
        if not self.is_master:
            return
        if self._checkpoint is not None:
            self._model.visualize(self._checkpoint.model_dir)
        tf.get_logger().info(
            "Number of model parameters: %d", self._model.count_params()
        )
        tf.get_logger().info(
            "Number of model weights: %d (trainable = %d, non trainable = %d)",
            len(self._model.weights),
            len(self._model.trainable_weights),
            len(self._model.non_trainable_weights),
        )

    def _should_record_summaries(self, accum_steps, report_steps):
        """Returns a boolean tensor to be used in tf.summary.record_if."""
        if report_steps is None or not self.is_master:
            return False
        record_summaries = tf.equal(self._optimizer.iterations % report_steps, 0)
        if accum_steps > 1:
            record_summaries = tf.logical_and(
                record_summaries, tf.equal(self._gradient_accumulator.step, 0)
            )
        return record_summaries

    def _compute_gradients(self, source, target, accum_steps, report_steps):
        """Computes the gradient of a training example."""
        record_summaries = self._should_record_summaries(accum_steps, report_steps)
        with tf.summary.record_if(record_summaries):
            reported_loss, gradients = self._model.compute_gradients(
                source,
                target,
                self._optimizer,
                loss_scale=accum_steps * self.num_replicas,
            )
            self._training_stats.update_on_example(source, target)
            _summarize_gradients(gradients, record_summaries)
        return reported_loss, gradients

    def _apply_gradients(self, gradients):
        """Applies the gradients."""
        self._optimizer.apply_gradients(
            list(zip(gradients, self._model.trainable_variables))
        )

    def _forward(self, source, target, accum_steps=1, report_steps=None):
        """Forwards a training example and accumulates the gradients."""
        loss, gradients = self._compute_gradients(
            source,
            target,
            accum_steps,
            report_steps,
        )
        if accum_steps > 1:
            self._gradient_accumulator(gradients)
        else:
            self._apply_gradients(gradients)
        return loss

    def _step(self):
        """Applies gradients and resets accumulation."""
        self._apply_gradients(self._gradient_accumulator.gradients)
        self._gradient_accumulator.reset()

    def _update_moving_average(self, moving_average):
        """Updates the moving average of variables."""
        moving_average.update()

    def _broadcast_variables(self):
        """Broadcasts variables to other replicas, if required."""
        return

    def _all_reduce_sum(self, value):
        """Reduces the value across all replicas."""
        return value


class HorovodTrainer(Trainer):
    """Trainer compatible with Horovod distributed training."""

    def __init__(self, model, optimizer, hvd, checkpoint=None):
        """Initializes the Horovod trainer.

        Args:
          model: A :class:`opennmt.models.Model` instance to train.
          optimizer: A ``tf.keras.optimizers.Optimizer`` instance.
          hvd: The global Horovod object.
          checkpoint: A :class:`opennmt.utils.checkpoint.Checkpoint` instance. If
            not set, no checkpoints will be saved.
        """
        super().__init__(model, optimizer, checkpoint=checkpoint)
        self._hvd = hvd

    @property
    def is_master(self):
        return self._hvd.rank() == 0

    @property
    def num_replicas(self):
        return self._hvd.size()

    def _finalize_dataset(self, dataset):
        if callable(dataset):
            dataset = dataset(
                tf.distribute.InputContext(
                    num_input_pipelines=self._hvd.size(),
                    input_pipeline_id=self._hvd.rank(),
                    num_replicas_in_sync=self._hvd.size(),
                )
            )
        return dataset

    def _apply_gradients(self, gradients):
        return super()._apply_gradients(map(self._all_reduce_sum, gradients))

    def _broadcast_variables(self):
        self._hvd.broadcast_variables(self._model.variables, root_rank=0)
        self._hvd.broadcast_variables(self._optimizer.variables(), root_rank=0)

    def _all_reduce_sum(self, value):
        return self._hvd.allreduce(value, op=self._hvd.Sum)


class MirroredStrategyTrainer(Trainer):
    """Trainer based on ``tf.distribute.MirroredStrategy`` for local multi-GPU training."""

    def __init__(self, model, optimizer, checkpoint=None, devices=None):
        """Initializes the MirroredStrategy trainer.

        Args:
          model: A :class:`opennmt.models.Model` instance to train.
          optimizer: A ``tf.keras.optimizers.Optimizer`` instance.
          checkpoint: A :class:`opennmt.utils.checkpoint.Checkpoint` instance. If
            not set, no checkpoints will be saved.
          devices: List of device strings to use for training. If not set, all
            visible GPUs are used.
        """
        super().__init__(model, optimizer, checkpoint=checkpoint)
        self._strategy = tf.distribute.MirroredStrategy(devices=devices)
        with self._strategy.scope():
            # Create some variables under the strategy scope.
            _ = self._optimizer.iterations

    @property
    def num_replicas(self):
        return self._strategy.num_replicas_in_sync

    def _finalize_dataset(self, dataset):
        # We prefer not to use experimental_distribute_dataset here because it
        # sometimes fails to split the batches (noticed with tokens batch type).
        dataset_fn = dataset if callable(dataset) else lambda _: dataset
        return self._strategy.distribute_datasets_from_function(dataset_fn)

    def _forward(self, source, target, accum_steps=1, report_steps=None):
        per_replica_loss = self._strategy.run(
            super()._forward,
            args=(source, target),
            kwargs=dict(accum_steps=accum_steps, report_steps=report_steps),
        )
        # TODO: this reduction could be delayed until _step is called.
        return self._strategy.reduce(tf.distribute.ReduceOp.SUM, per_replica_loss, None)

    def _step(self):
        self._strategy.run(super()._step)

    def _update_moving_average(self, moving_average):
        with self._strategy.scope():
            super()._update_moving_average(moving_average)


def _summarize_gradients(gradients, should_record):
    # Only compute the gradients global norm when the value is actually recorded.
    if isinstance(should_record, bool) and not should_record:
        return
    tf.summary.scalar(
        "gradients/global_norm",
        tf.cond(
            should_record,
            true_fn=lambda: tf.linalg.global_norm(gradients),
            false_fn=lambda: tf.constant(0, dtype=gradients[0].dtype),
        ),
    )


class MovingAverage(object):
    """Object holding an exponential moving average of variables."""

    def __init__(self, variables, step, decay=0.9999):
        """Initializes the moving average object.

        Args:
          variables: The list of variable for which to maintain a moving average.
          step: The training step counter as a ``tf.Variable``.
          decay: The decay rate of the exponential moving average. Usually close to
            1, e.g. 0.9999, see the complete formula on
            https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage.

        Raises:
          TypeError: is :obj:`step` is not a ``tf.Variable``.
        """
        if not isinstance(step, tf.Variable):
            raise TypeError("step should be a tf.Variable")
        if decay < 0.9 or decay > 1:
            tf.get_logger().warning(
                "Moving average decay should be close to 1 (e.g. 0.9999) but you "
                "passed %f, is it correct? See https://www.tensorflow.org/api_docs"
                "/python/tf/train/ExponentialMovingAverage for details about the "
                "formula and recommended decay values."
            )
        self._ema = tf.train.ExponentialMovingAverage(decay, num_updates=step)
        self._variables = variables

    @tf.function
    def update(self):
        """Updates the moving average of the variables."""
        self._ema.apply(self._variables)

    @contextlib.contextmanager
    def shadow_variables(self):
        """Returns a context manager that assigns the variables to their moving
        average value on enter and restores the previous value on exit.

        Returns:
          A context manager.
        """
        previous_values = []
        for variable in self._variables:
            previous_values.append(variable.value())
            variable.assign(self._ema.average(variable))
        yield
        for previous_value, variable in zip(previous_values, self._variables):
            variable.assign(previous_value)


class TrainingStats:
    """Aggregate and summarize training statistics."""

    def __init__(self, model, optimizer, reduce_fn=None, warmup_steps=2):
        """Initializes the statistics.

        Args:
          model: The model.
          optimizer: The optimizer.
          reduce_fn: In case of distributed training, a function to sum reduce
            distributed values.
          warmup_steps: Throughput values are ignored for this many steps at the
            start of the training.
        """
        self._model = model
        self._optimizer = optimizer
        self._reduce_fn = reduce_fn
        self._warmup_steps = warmup_steps
        self._words_counters = {}
        self._num_updates = 0
        self._average_loss = 0
        self._last_loss = None
        self._last_step = optimizer.iterations.numpy()
        self._last_logged_step = self._last_step
        self._last_logged_time = time.time()
        self._num_tokens = collections.defaultdict(int)
        self._oov_tokens = collections.defaultdict(lambda: collections.defaultdict(int))

    def update_on_example(self, source, target):
        """Updates the training statistics on a new training example.

        This may be called within a tf.function.

        Args:
          source: A dictionary of source features.
          target: A dictionary of target features.
        """
        self._update_words_counter("source", source)
        self._record_oov_tokens("source", source, self._model.features_inputter)
        if not self._model.unsupervised:
            self._update_words_counter("target", target)
            self._record_oov_tokens("target", target, self._model.labels_inputter)

    def update_on_step(self, step, loss):
        """Updates the training statistics on a new training step.

        Args:
          step: The current training step.
          loss: The loss for this step.
        """
        # Convert Numpy or Tensor values to Python.
        step = int(step)
        loss = float(loss)

        self._last_step = step
        self._last_loss = loss
        self._average_loss = (self._average_loss * self._num_updates + loss) / (
            self._num_updates + 1
        )

        if self._num_updates < self._warmup_steps:
            self.reset_throughput()
        self._num_updates += 1

    def get_last_summary(self):
        """Returns a summary of the training since the last log.

        Returns:
          A dictionary containing various statistics.
        """
        elapsed_time = time.time() - self._last_logged_time
        return {
            "learning_rate": self._get_learning_rate(),
            "loss": self._last_loss,
            "step": self._last_step,
            "steps_per_sec": (self._last_step - self._last_logged_step) / elapsed_time,
            "words_per_sec": {
                name: int(value.numpy() / elapsed_time)
                for name, value in self._get_words_counters().items()
            },
        }

    def get_global_summary(self):
        """Returns a summary of the training since the beginning.

        Returns:
          A dictionary containing various statistics.
        """
        return {
            "average_loss": self._average_loss,
            "last_learning_rate": self._get_learning_rate(),
            "last_loss": self._last_loss,
            "last_step": self._last_step,
            "num_steps": self._num_updates,
        }

    def log(self, is_master=True):
        """Logs the last training statistics.

        Args:
          is_master: Whether this process is the master worker or not.
        """

        # Only the master should log the training statistics but we build the
        # summary on all workers since it may reduce distributed values.
        summary = self.get_last_summary()

        if not is_master:
            return

        tf.summary.scalar(
            "steps_per_sec",
            summary["steps_per_sec"],
            description="Training steps per second",
        )
        steps_per_sec_fmt = "steps/s = %0.2f" % summary["steps_per_sec"]

        words_per_sec_fmt = []
        for name, avg in summary["words_per_sec"].items():
            tf.summary.scalar(
                "words_per_sec/%s" % name,
                avg,
                description="%s words per second" % name.capitalize(),
            )
            words_per_sec_fmt.append("%s words/s = %d" % (name, avg))

        tf.get_logger().info(
            "Step = %d ; %s ; Learning rate = %f ; Loss = %f",
            summary["step"],
            ", ".join([steps_per_sec_fmt] + list(sorted(words_per_sec_fmt))),
            summary["learning_rate"],
            summary["loss"],
        )
        tf.summary.scalar("loss", summary["loss"], description="Training loss")
        tf.summary.scalar(
            "optim/learning_rate", summary["learning_rate"], description="Learning rate"
        )

    def log_final(self, is_master=True):
        """Outputs the final log."""
        if not is_master:
            return

        for name, oov_tokens in self._oov_tokens.items():
            num_oov_tokens = sum(oov_tokens.values())
            if num_oov_tokens > 0:
                num_tokens = self._num_tokens[name]
                tf.get_logger().warning(
                    "%.3f%% of %s tokens are out of vocabulary (%d out of %d tokens)",
                    (num_oov_tokens / num_tokens) * 100,
                    name,
                    num_oov_tokens,
                    num_tokens,
                )
                most_frequent_oov_tokens = (
                    "%s (%.1f%%)" % (oov_token, (count / num_oov_tokens) * 100)
                    for oov_token, count in sorted(
                        oov_tokens.items(),
                        key=lambda x: x[1],
                        reverse=True,
                    )
                )
                most_frequent_oov_tokens = list(most_frequent_oov_tokens)[:10]
                tf.get_logger().info(
                    "The %d most frequent out of vocabulary %s tokens are: %s",
                    len(most_frequent_oov_tokens),
                    name,
                    "; ".join(most_frequent_oov_tokens),
                )

    def reset_throughput(self):
        """Resets the accumulated values since the last log."""
        self._reset_words_counters()
        self._last_logged_step = self._last_step
        self._last_logged_time = time.time()

    def _get_learning_rate(self):
        learning_rate = self._optimizer.learning_rate
        if isinstance(learning_rate, tf.optimizers.schedules.LearningRateSchedule):
            learning_rate = learning_rate(self._last_step)
        return float(learning_rate)

    def _record_oov_tokens(self, name, features, inputter):
        if not isinstance(inputter, text_inputter.WordEmbedder):
            return

        def _record(num_tokens, oov_tokens):
            self._num_tokens[name] += int(num_tokens)
            all_oov_tokens = self._oov_tokens[name]
            for oov_token in oov_tokens.flatten():
                all_oov_tokens[oov_token.decode("utf-8")] += 1

        num_tokens = tf.reduce_sum(
            inputter.get_length(features, ignore_special_tokens=True)
        )
        oov_tokens = inputter.get_oov_tokens(features)

        tf.numpy_function(_record, [num_tokens, oov_tokens], [])

    def _update_words_counter(self, name, features):
        """Accumulates the number of source and target tokens to report throughput."""
        length = features.get("length")
        if length is None:
            return
        num_words = tf.reduce_sum(length)
        counter = self._words_counters.get(name)
        if counter is None:
            counter = tf.Variable(
                tf.constant(0, dtype=tf.int64),
                trainable=False,
                synchronization=tf.VariableSynchronization.ON_READ,
                aggregation=tf.VariableAggregation.SUM,
            )
            self._words_counters[name] = counter
        counter.assign_add(tf.cast(num_words, tf.int64))

    @tf.function
    def _get_words_counters(self):
        """Returns the accumulated words counters.

        Returns:
          A dictionary mapping a counter name to a value.
        """
        counters = {}
        for name, counter in self._words_counters.items():
            counter = counter.read_value()
            if self._reduce_fn is not None:
                counter = self._reduce_fn(counter)
            counters[name] = counter
        return counters

    @tf.function
    def _reset_words_counters(self):
        """Resets the accumulated words counters."""
        for counter in self._words_counters.values():
            counter.assign(tf.constant(0, dtype=tf.int64))



"""OpenNMT-tf version."""

__version__ = "2.25.0"

INCLUSIVE_MIN_TF_VERSION = "2.4.0"
EXCLUSIVE_MAX_TF_VERSION = "2.9.0"


def _check_tf_version():
    import warnings

    from distutils.version import LooseVersion

    import tensorflow as tf

    if (
        LooseVersion(INCLUSIVE_MIN_TF_VERSION)
        <= LooseVersion(tf.__version__)
        < LooseVersion(EXCLUSIVE_MAX_TF_VERSION)
    ):
        return

    warnings.warn(
        "OpenNMT-tf supports TensorFlow versions %s (included) to %s (excluded), "
        "but you have TensorFlow %s installed. Some features might not work properly."
        % (INCLUSIVE_MIN_TF_VERSION, EXCLUSIVE_MAX_TF_VERSION, tf.__version__),
        UserWarning,
    )



"""Defines functions related to configuration files."""

import copy
import importlib
import os
import sys

import tensorflow as tf
import yaml

from opennmt.models import catalog
from opennmt.optimizers import utils as optimizers_lib
from opennmt.schedules import lr_schedules as schedules_lib
from opennmt.utils.misc import merge_dict

MODEL_DESCRIPTION_FILENAME = "model_description.py"


def load_model_module(path):
    """Loads a model configuration file.

    Args:
      path: The relative path to the configuration file.

    Returns:
      A Python module.

    Raises:
      ValueError: if :obj:`path` is invalid.
      ImportError: if the module in :obj:`path` does not define a model.
    """
    if not os.path.exists(path):
        raise ValueError("Model configuration not found in %s" % path)
    dirname, filename = os.path.split(path)
    module_name, _ = os.path.splitext(filename)
    sys.path.insert(0, os.path.abspath(dirname))
    module = importlib.import_module(module_name)
    sys.path.pop(0)

    if not hasattr(module, "model"):
        raise ImportError("No model defined in {}".format(path))

    return module


def load_model_from_file(path, as_builder=False):
    """Loads a model from a configuration file.

    Args:
      path: The relative path to the configuration file.
      as_builder: If ``True``, return a callable building the model on call.

    Returns:
      A :class:`opennmt.models.Model` instance or a callable returning such
      instance.
    """
    module = load_model_module(path)
    model = module.model
    if not as_builder:
        model = model()
    del sys.path_importer_cache[os.path.dirname(module.__file__)]
    del sys.modules[module.__name__]
    return model


def load_model_from_catalog(name, as_builder=False):
    """Loads a model from the catalog.

    Args:
      name: The model name.
      as_builder: If ``True``, return a callable building the model on call.

    Returns:
      A :class:`opennmt.models.Model` instance or a callable returning such
      instance.

    Raises:
      ValueError: if the model :obj:`name` does not exist in the model catalog.
    """
    return catalog.get_model_from_catalog(name, as_builder=as_builder)


def load_model(
    model_dir, model_file=None, model_name=None, serialize_model=True, as_builder=False
):
    """Loads the model from the catalog or a definition file.

    Args:
      model_dir: The model directory.
      model_file: An optional model configuration.
        Mutually exclusive with :obj:`model_name`.
      model_name: An optional model name from the catalog.
        Mutually exclusive with :obj:`model_file`.
      serialize_model: Serialize the model definition in the model directory to
        make it optional for future runs.
      as_builder: If ``True``, return a callable building the model on call.

    Returns:
      A :class:`opennmt.models.Model` instance or a callable returning such
      instance.

    Raises:
      ValueError: if both :obj:`model_file` and :obj:`model_name` are set.
    """
    if model_file and model_name:
        raise ValueError("only one of model_file and model_name should be set")
    model_description_path = os.path.join(model_dir, MODEL_DESCRIPTION_FILENAME)

    if model_file:
        model = load_model_from_file(model_file, as_builder=as_builder)
        if serialize_model:
            tf.io.gfile.copy(model_file, model_description_path, overwrite=True)
    elif model_name:
        model = load_model_from_catalog(model_name, as_builder=as_builder)
        if serialize_model:
            with tf.io.gfile.GFile(
                model_description_path, mode="w"
            ) as model_description_file:
                model_description_file.write(
                    "from opennmt import models\n"
                    'model = lambda: models.get_model_from_catalog("%s")\n' % model_name
                )
    elif tf.io.gfile.exists(model_description_path):
        tf.get_logger().info(
            "Loading model description from %s", model_description_path
        )
        model = load_model_from_file(model_description_path, as_builder=as_builder)
    else:
        raise RuntimeError(
            "A model configuration is required: you probably need to "
            "set --model or --model_type on the command line."
        )

    return model


def load_config(config_paths, config=None):
    """Loads YAML configuration files.

    Args:
      config_paths: A list of configuration files that will be merged to a single
        configuration. The rightmost configuration takes priority.
      config: A (possibly non empty) config dictionary to fill.

    Returns:
      The configuration as Python dictionary.
    """
    if config is None:
        config = {}

    for config_path in config_paths:
        with tf.io.gfile.GFile(config_path) as config_file:
            subconfig = yaml.safe_load(config_file.read())
            # Add or update section in main configuration.
            merge_config(config, subconfig)

    return config


def merge_config(a, b):
    """Merges the configuration :obj:`b` into the configuration :obj:`a`."""
    override_keys = {"optimizer_params", "decay_params"}
    return merge_dict(a, b, override_keys=override_keys)


def _map_config_values(config, fn):
    """Applies the function ``fn`` on each value of the configuration."""
    if isinstance(config, dict):
        return {key: _map_config_values(value, fn) for key, value in config.items()}
    elif isinstance(config, list):
        return [_map_config_values(elem, fn) for elem in config]
    else:
        return fn(config)


def try_prefix_paths(prefix, config):
    """Recursively tries to prefix paths in the configuration.

    The path is unchanged if the prefixed path does not exist.

    Args:
      prefix: The prefix to apply.
      config: The configuration.

    Returns:
      The updated configuration.
    """

    def _fn(path):
        if isinstance(path, str):
            new_path = os.path.join(prefix, path)
            if tf.io.gfile.exists(new_path):
                return new_path
        return path

    return _map_config_values(config, _fn)


def convert_to_v2_config(v1_config):
    """Converts a V1 configuration to its V2 equivalent.

    Args:
      v1_config: The V1 configuration.

    Returns:
      The V2 configuration.

    Raises:
      ValueError: if the conversion can not be done automatically.
    """
    config = copy.deepcopy(v1_config)

    _convert_to_v2_params(config)

    data_config = config.get("data")
    if data_config:
        # Covers most of seq2seq models in the catalog.
        _rename_opt(data_config, "source_words_vocabulary", "source_vocabulary")
        _rename_opt(data_config, "target_words_vocabulary", "target_vocabulary")

    for section_name in ("train", "eval", "infer", "score", "params"):
        section = config.get(section_name)
        if section is None:
            continue

        _delete_opt(section, "num_threads")
        _delete_opt(section, "prefetch_buffer_size")
        _rename_opt(section, "bucket_width", "length_bucket_width")
        if section_name == "train":
            _rename_opt(section, "train_steps", "max_step")
            _delete_opt(section, "save_checkpoints_secs")
        elif section_name == "eval":
            _delete_opt(section, "eval_delay")
            _delete_opt(section, "exporters")

        # Remove empty sections.
        if not section:
            config.pop(section_name)

    return config


def _convert_to_v2_params(config):
    params = config.get("params")
    if not params:
        return
    if "freeze_variables" in params:
        raise ValueError(
            "params/freeze_variables should be manually converted to "
            "params/freeze_layers"
        )

    _convert_to_v2_optimizer(params)
    _convert_to_v2_lr_schedules(params)
    _convert_to_v2_step_accumulation(params, config)
    _delete_opt(params, "param_init")
    _delete_opt(params, "loss_scale")
    _delete_opt(params, "horovod")
    _delete_opt(params, "maximum_learning_rate")
    _rename_opt(params, "maximum_iterations", "maximum_decoding_length")

    clip_gradients = _delete_opt(params, "clip_gradients")
    if clip_gradients:
        optimizer_params = params.setdefault("optimizer_params", {})
        optimizer_params["clipnorm"] = clip_gradients

    weight_decay = _delete_opt(params, "weight_decay")
    if weight_decay:
        optimizer_params = params.setdefault("optimizer_params", {})
        optimizer_params["weight_decay"] = weight_decay


# Only covering the most common optimizers.
_V1_OPTIMIZER_MAP = {
    "AdamOptimizer": "Adam",
    "GradientDescentOptimizer": "SGD",
    "LazyAdamOptimizer": "LazyAdam",
}


def _convert_to_v2_optimizer(params):
    optimizer = params.get("optimizer")
    if optimizer:
        try:
            # Check if the optimizer exists in V2.
            optimizers_lib.get_optimizer_class(optimizer)
        except ValueError as e:
            v2_optimizer = _V1_OPTIMIZER_MAP.get(optimizer)
            if not v2_optimizer:
                raise ValueError(
                    "params/optimizer should be manually converted: no registered "
                    "conversion for optimizer %s" % optimizer
                ) from e
            params["optimizer"] = v2_optimizer

    optimizer_params = params.get("optimizer_params")
    if optimizer_params:
        _rename_opt(optimizer_params, "beta1", "beta_1")
        _rename_opt(optimizer_params, "beta2", "beta_2")


def _convert_to_v2_lr_schedules(params):
    decay_type = params.get("decay_type")
    if not decay_type:
        return
    try:
        # Check if the learning rate schedule exists in V2.
        schedules_lib.get_lr_schedule_class(decay_type)
    except ValueError as e:
        if decay_type.startswith("noam_decay"):
            params["decay_type"] = "NoamDecay"
            if "decay_params" not in params:
                model_dim = _delete_opt(params, "decay_rate")
                warmup_steps = _delete_opt(params, "decay_steps")
                params["decay_params"] = dict(
                    model_dim=model_dim, warmup_steps=warmup_steps
                )
        else:
            raise ValueError(
                "params/decay_type should be manually converted: no registered "
                "conversion for decay type %s" % decay_type
            ) from e


def _convert_to_v2_step_accumulation(params, config):
    # Try to upgrade step accumulation to train/effective_batch_size.
    decay_step_duration = _delete_opt(params, "decay_step_duration") or 1
    gradients_accum = _delete_opt(params, "gradients_accum") or 1
    accum_steps = max(decay_step_duration, gradients_accum)
    if accum_steps > 1:
        train_config = config.setdefault("train", {})
        batch_size = train_config.get("batch_size")
        if batch_size is not None and batch_size > 0:
            train_config["effective_batch_size"] = batch_size * accum_steps
        else:
            raise ValueError(
                "params/decay_step_duration and params/gradients_accum "
                "should be manually converted to train/effective_batch_size"
            )


def _delete_opt(config, name):
    return config.pop(name, None)


def _rename_opt(config, name, new_name):
    value = _delete_opt(config, name)
    if value is not None:
        config[new_name] = value
    return value



"""Functions for compatibility with different TensorFlow versions."""

import tensorflow as tf


def tf_supports(symbol):
    """Returns ``True`` if TensorFlow defines :obj:`symbol`."""
    return _string_to_tf_symbol(symbol) is not None


def tf_any(*symbols):
    """Returns the first supported symbol."""
    for symbol in symbols:
        module = _string_to_tf_symbol(symbol)
        if module is not None:
            return module
    return None


def _string_to_tf_symbol(symbol):
    modules = symbol.split(".")
    namespace = tf
    for module in modules:
        namespace = getattr(namespace, module, None)
        if namespace is None:
            return None
    return namespace



"""SentencePiece tokenizer."""

import tensorflow as tf
import tensorflow_text as tft

from opennmt.tokenizers import tokenizer


@tokenizer.register_tokenizer
class SentencePieceTokenizer(tokenizer.TensorFlowTokenizer):
    """In-graph SentencePiece tokenizer using
    ``tensorflow_text.SentencepieceTokenizer``.
    """

    def __init__(self, model, nbest_size=0, alpha=1.0):
        """Initializes the tokenizer.

        Args:
          model: Path to the SentencePiece model.
          nbest_size: Number of candidates to sample from (disabled during inference).
          alpha: Smoothing parameter for the sampling.
        """
        super().__init__()
        self._nbest_size = nbest_size
        with tf.io.gfile.GFile(model, "rb") as model_file:
            self._tokenizer = tft.SentencepieceTokenizer(
                model=model_file.read(),
                out_type=tf.string,
                nbest_size=nbest_size,
                alpha=alpha,
            )

    def _tokenize_tensor(self, text, training):
        if not training:
            self._tokenizer.nbest_size = 0
        tokens = self._tokenizer.tokenize(text)
        if not training:
            self._tokenizer.nbest_size = self._nbest_size
        return tokens

    def _detokenize_tensor(self, tokens):
        return self._tokenizer.detokenize(tokens)



"""Various utility functions to use throughout the project."""

import collections
import copy
import functools
import heapq
import io
import os
import sys

import numpy as np
import tensorflow as tf

from tensorflow.python.training.tracking import graph_view


def get_devices(count=1, fallback_to_cpu=True):
    """Gets devices.

    Args:
      count: The number of devices to get.
      fallback_to_cpu: If ``True``, return CPU devices if no GPU is available.

    Returns:
      A list of device names.

    Raises:
      ValueError: if :obj:`count` is greater than the number of visible devices.
    """
    device_type = "GPU"
    devices = tf.config.list_logical_devices(device_type=device_type)
    if not devices and fallback_to_cpu:
        tf.get_logger().warning("No GPU is detected, falling back to CPU")
        device_type = "CPU"
        devices = tf.config.list_logical_devices(device_type=device_type)
    if len(devices) < count:
        raise ValueError(
            "Requested %d %s devices but %d %s %s visible"
            % (
                count,
                device_type,
                len(devices),
                device_type,
                "is" if len(devices) == 1 else "are",
            )
        )
    return devices[0:count]


def enable_mixed_precision(force=False):
    """Globally enables mixed precision if the detected hardware supports it.

    Args:
      force: Set ``True`` to force mixed precision mode even if the hardware
        does not support it.

    Returns:
      A boolean to indicate whether mixed precision was enabled or not.
    """
    if not force:
        gpu_devices = tf.config.get_visible_devices("GPU")
        if not gpu_devices:
            tf.get_logger().warning("Mixed precision not enabled: no GPU is detected")
            return False

        gpu_details = tf.config.experimental.get_device_details(gpu_devices[0])
        compute_capability = gpu_details.get("compute_capability")
        if compute_capability is None:
            tf.get_logger().warning(
                "Mixed precision not enabled: a NVIDIA GPU is required"
            )
            return False
        if compute_capability < (7, 0):
            tf.get_logger().warning(
                "Mixed precision not enabled: a NVIDIA GPU with compute "
                "capability 7.0 or above is required, but the detected GPU "
                "has compute capability %d.%d" % compute_capability
            )
            return False

    tf.keras.mixed_precision.set_global_policy("mixed_float16")
    return True


def disable_mixed_precision():
    """Globally disables mixed precision."""
    tf.keras.mixed_precision.set_global_policy("float32")


def mixed_precision_enabled():
    """Returns ``True`` if mixed precision is enabled."""
    policy = tf.keras.mixed_precision.global_policy()
    return "float16" in policy.name


def get_variables_name_mapping(root, root_key=None):
    """Returns mapping between variables and their name in the object-based
    representation.

    Args:
      root: The root layer.
      root_key: Key that was used to save :obj:`root`, if any.

    Returns:
      A dict mapping names to variables.
    """
    # TODO: find a way to implement this function using public APIs.
    names_to_variables = {}
    _, path_to_root = graph_view.ObjectGraphView(root)._breadth_first_traversal()
    for path in path_to_root.values():
        if not path:
            continue
        variable = path[-1].ref
        if not isinstance(variable, tf.Variable):
            continue
        name = "%s/%s" % (
            "/".join(field.name for field in path),
            ".ATTRIBUTES/VARIABLE_VALUE",
        )
        if root_key is not None:
            name = "%s/%s" % (root_key, name)
        names_to_variables[name] = variable
    return names_to_variables


def get_variable_name(variable, root, model_key="model"):
    """Gets the variable name in the object-based representation."""
    names_to_variables = get_variables_name_mapping(root, root_key=model_key)
    for name, var in names_to_variables.items():
        if var is variable:
            return name
    return None


def print_as_bytes(text, stream=None):
    """Prints a string as bytes to not rely on :obj:`stream` default encoding.

    Args:
      text: The text to print.
      stream: The stream to print to (``sys.stdout`` if not set).
    """
    if stream is None:
        stream = sys.stdout
    write_buffer = stream.buffer if hasattr(stream, "buffer") else stream
    write_buffer.write(tf.compat.as_bytes(text))
    write_buffer.write(b"\n")
    stream.flush()


def format_translation_output(
    sentence, score=None, token_level_scores=None, attention=None, alignment_type=None
):
    """Formats a translation output with possibly scores, alignments, etc., e.g:

    1.123214 ||| Hello world ||| 0.30907777 0.030488174 ||| 0-0 1-1

    Args:
      sentence: The translation to output.
      score: If set, attach the score.
      token_level_scores: If set, attach the token level scores.
      attention: The attention vector.
      alignment_type: The type of alignments to format (can be: "hard", "soft").
    """
    if score is not None:
        sentence = "%f ||| %s" % (score, sentence)
    if token_level_scores is not None:
        scores_str = " ".join("%f" % s for s in token_level_scores)
        sentence = "%s ||| %s" % (sentence, scores_str)
    if attention is not None and alignment_type is not None:
        if alignment_type == "hard":
            source_indices = np.argmax(attention, axis=-1)
            target_indices = range(attention.shape[0])
            pairs = (
                "%d-%d" % (src, tgt) for src, tgt in zip(source_indices, target_indices)
            )
            sentence = "%s ||| %s" % (sentence, " ".join(pairs))
        elif alignment_type == "soft":
            vectors = []
            for vector in attention:
                vectors.append(" ".join("%.6f" % value for value in vector))
            sentence = "%s ||| %s" % (sentence, " ; ".join(vectors))
        else:
            raise ValueError("Invalid alignment type %s" % alignment_type)
    return sentence


def item_or_tuple(x):
    """Returns :obj:`x` as a tuple or its single element."""
    x = tuple(x)
    if len(x) == 1:
        return x[0]
    else:
        return x


def count_lines(filename, buffer_size=65536):
    """Returns the number of lines of the file :obj:`filename`."""
    with tf.io.gfile.GFile(filename, mode="rb") as f:
        num_lines = 0
        while True:
            data = f.read(buffer_size)
            if not data:
                return num_lines
            num_lines += data.count(b"\n")


def is_gzip_file(filename):
    """Returns ``True`` if :obj:`filename` is a GZIP file."""
    return filename.endswith(".gz")


def shape_list(x):
    """Return list of dims, statically where possible."""
    x = tf.convert_to_tensor(x)
    if tf.executing_eagerly():
        return x.shape.as_list()

    # If unknown rank, return dynamic shape
    if x.shape.dims is None:
        return tf.shape(x)

    static = x.shape.as_list()
    shape = tf.shape(x)

    ret = []
    for i, _ in enumerate(static):
        dim = static[i]
        if dim is None:
            dim = shape[i]
        ret.append(dim)
    return ret


def index_structure(structure, path, path_separator="/"):
    """Follows :obj:`path` in a nested structure of objects, lists, and dicts."""
    keys = path.split(path_separator)
    for i, key in enumerate(keys):
        current_path = "%s%s" % (path_separator, path_separator.join(keys[:i]))
        if isinstance(structure, list):
            try:
                index = int(key)
            except ValueError as e:
                raise ValueError(
                    "Object referenced by path '%s' is a list, but got non "
                    "integer index '%s'" % (current_path, key)
                ) from e
            if index < 0 or index >= len(structure):
                raise ValueError(
                    "List referenced by path '%s' has length %d, but got "
                    "out of range index %d" % (current_path, len(structure), index)
                )
            structure = structure[index]
        elif isinstance(structure, dict):
            structure = structure.get(key)
            if structure is None:
                raise ValueError(
                    "Dictionary referenced by path '%s' does not have the "
                    "key '%s'" % (current_path, key)
                )
        else:
            structure = getattr(structure, key, None)
            if structure is None:
                raise ValueError(
                    "Object referenced by path '%s' does not have the "
                    "attribute '%s'" % (current_path, key)
                )
    return structure


def clone_layer(layer):
    """Clones a layer."""
    return copy.deepcopy(layer)


def set_dropout(root_layer, dropout):
    """Overrides all dropout values in :obj:`root_layer` and its descendants.

    Args:
      dropout: The dropout value to set.

    Raises:
      ValueError: if :obj:`root_layer` is not a ``tf.Module``.
    """
    if not isinstance(root_layer, tf.Module):
        raise ValueError("Layer should be a tf.Module")
    for layer in (root_layer,) + root_layer.submodules:
        for attr, value in layer.__dict__.copy().items():
            if isinstance(value, tf.keras.layers.Dropout):
                value.rate = dropout
            elif "dropout" in attr:
                setattr(layer, attr, dropout)


def describe_layer(layer, name=None):
    """Returns a PyTorch-style description of the layer, for information or debug."""
    with io.StringIO() as output:
        _describe_layer(output, layer, name=name)
        return output.getvalue()


def _describe_layer(output, layer, name=None, indent=""):
    if indent:
        output.write(indent)
    if name:
        output.write("(%s): " % name)
    output.write("%s(" % layer.__class__.__name__)
    if isinstance(layer, list):
        children = list(enumerate(layer))
    else:
        children = _get_direct_children(layer)
    if not children:
        units = getattr(layer, "units", None)
        if units is not None:
            output.write("%d" % units)
    else:
        output.write("\n")
        for attr_name, child in children:
            _describe_layer(output, child, name=str(attr_name), indent=indent + "  ")
        if indent:
            output.write(indent)
    output.write(")\n")


def _get_direct_children(layer):
    children = []
    for name, attr in layer.__dict__.items():
        if name.startswith("_"):
            continue
        if isinstance(attr, tf.Module) or (
            isinstance(attr, list) and attr and isinstance(attr[0], tf.Module)
        ):
            children.append((name, attr))
    return children


def extract_batches(tensors):
    """Returns a generator to iterate on each batch of a Numpy array or dict of
    Numpy arrays."""
    if not isinstance(tensors, dict):
        for tensor in tensors:
            yield tensor
    else:
        batch_size = None
        for value in tensors.values():
            batch_size = batch_size or value.shape[0]
        for b in range(batch_size):
            yield {key: value[b] for key, value in tensors.items()}


def extract_prefixed_keys(dictionary, prefix):
    """Returns a dictionary with all keys from :obj:`dictionary` that are prefixed
    with :obj:`prefix`.
    """
    sub_dict = {}
    for key, value in dictionary.items():
        if key.startswith(prefix):
            original_key = key[len(prefix) :]
            sub_dict[original_key] = value
    return sub_dict


def extract_suffixed_keys(dictionary, suffix):
    """Returns a dictionary with all keys from :obj:`dictionary` that are suffixed
    with :obj:`suffix`.
    """
    sub_dict = {}
    for key, value in dictionary.items():
        if key.endswith(suffix):
            original_key = key[: -len(suffix)]
            sub_dict[original_key] = value
    return sub_dict


def merge_dict(dict1, dict2, override_keys=None):
    """Merges :obj:`dict2` into :obj:`dict1`.

    Args:
      dict1: The base dictionary.
      dict2: The dictionary to merge.
      override_keys: The values associated with these keys are overridden instead
        of merged.

    Returns:
      The merged dictionary :obj:`dict1`.
    """
    for key, value in dict2.items():
        if isinstance(value, dict) and (
            override_keys is None or key not in override_keys
        ):
            dict1[key] = merge_dict(
                dict1.get(key, {}), value, override_keys=override_keys
            )
        else:
            dict1[key] = value
    return dict1


def read_summaries(event_dir, event_file_pattern="events.out.tfevents.*"):
    """Reads summaries from TensorFlow event files.

    Args:
      event_dir: Directory containing event files.
      event_file_pattern: The pattern to look for event files.

    Returns:
      A list of tuple (step, dict of summaries), sorted by step.
    """
    if not tf.io.gfile.exists(event_dir):
        return []
    summaries = collections.defaultdict(dict)
    for event_file in tf.io.gfile.glob(os.path.join(event_dir, event_file_pattern)):
        for event in tf.compat.v1.train.summary_iterator(event_file):
            if not event.HasField("summary"):
                continue
            for value in event.summary.value:
                tensor_proto = value.tensor
                tensor = tf.io.parse_tensor(
                    tensor_proto.SerializeToString(), tf.as_dtype(tensor_proto.dtype)
                )
                summaries[event.step][value.tag] = tf.get_static_value(tensor)
    return list(sorted(summaries.items(), key=lambda x: x[0]))


class OrderRestorer(object):
    """Helper class to restore out-of-order elements in order."""

    def __init__(self, index_fn, callback_fn):
        """Initializes this object.

        Args:
          index_fn: A callable mapping an element to a unique index.
          callback_fn: A callable taking an element that will be called in order.
        """
        self._index_fn = index_fn
        self._callback_fn = callback_fn
        self._next_index = 0
        self._elements = {}
        self._heap = []

    @property
    def buffer_size(self):
        """Number of elements waiting to be notified."""
        return len(self._heap)

    @property
    def next_index(self):
        """The next index to be notified."""
        return self._next_index

    def _try_notify(self):
        old_index = self._next_index
        while self._heap and self._heap[0] == self._next_index:
            index = heapq.heappop(self._heap)
            value = self._elements.pop(index)
            self._callback_fn(value)
            self._next_index += 1
        return self._next_index != old_index

    def push(self, x):
        """Push event :obj:`x`."""
        index = self._index_fn(x)
        if index is None:
            self._callback_fn(x)
            return True
        if index < self._next_index:
            raise ValueError("Event index %d was already notified" % index)
        self._elements[index] = x
        heapq.heappush(self._heap, index)
        return self._try_notify()


class ClassRegistry(object):
    """Helper class to create a registry of classes."""

    def __init__(self, base_class=None):
        """Initializes the class registry.

        Args:
          base_class: Ensure that classes added to this registry are a subclass of
            :obj:`base_class`.
        """
        self._base_class = base_class
        self._registry = {}

    @property
    def class_names(self):
        """Class names registered in this registry."""
        return set(self._registry.keys())

    def register(self, cls=None, name=None, alias=None):
        """Registers a class.

        Args:
          cls: The class to register. If not set, this method returns a decorator for
            registration.
          name: The class name. Defaults to ``cls.__name__``.
          alias: An optional alias or list of alias for this class.

        Returns:
          :obj:`cls` if set, else a class decorator.

        Raises:
          TypeError: if :obj:`cls` does not extend the expected base class.
          ValueError: if the class name is already registered.
        """
        if cls is None:
            return functools.partial(self.register, name=name, alias=alias)
        if self._base_class is not None and not issubclass(cls, self._base_class):
            raise TypeError(
                "Class %s does not extend %s"
                % (cls.__name__, self._base_class.__name__)
            )
        if name is None:
            name = cls.__name__
        self._register(cls, name)
        if alias is not None:
            if not isinstance(alias, (list, tuple)):
                alias = (alias,)
            for alias_name in alias:
                self._register(cls, alias_name)
        return cls

    def _register(self, cls, name):
        if name in self._registry:
            raise ValueError("Class name %s is already registered" % name)
        self._registry[name] = cls

    def get(self, name):
        """Returns the class with name :obj:`name` or ``None`` if it does not exist
        in the registry.
        """
        return self._registry.get(name)


class RelativeConfig(collections.abc.Mapping):
    """Helper class to lookup keys relative to a prefix."""

    def __init__(self, config, prefix=None, config_name=None):
        """Initializes the relative configuration.

        Args:
          config: The configuration.
          prefix: The prefix. Keys will be looked up relative to this prefix.
          config_name: The name of the configuration, mostly used to make error
            messages more explicit.
        """
        self._config = config
        self._prefix = prefix or ""
        self._config_name = config_name

    def __getitem__(self, relative_key):
        absolute_key = "%s%s" % (self._prefix, relative_key)
        value = self._config.get(absolute_key)
        if value is not None:
            return value
        value = self._config.get(relative_key)
        if value is not None:
            return value
        raise KeyError(
            "Missing field '%s' in the %sconfiguration"
            % (absolute_key, self._config_name + " " if self._config_name else "")
        )

    def __len__(self):
        return len(self._config)

    def __iter__(self):
        return iter(self._config)



"""Sequence tagger."""

import numpy as np
import tensorflow as tf
import tensorflow_addons as tfa

from opennmt import inputters
from opennmt.models.model import Model
from opennmt.utils.losses import cross_entropy_sequence_loss


class SequenceTagger(Model):
    """A sequence tagger."""

    def __init__(self, inputter, encoder, crf_decoding=False):
        """Initializes a sequence tagger.

        Args:
          inputter: A :class:`opennmt.inputters.Inputter` to process the
            input data.
          encoder: A :class:`opennmt.encoders.Encoder` to encode the input.
          crf_decoding: If ``True``, add a CRF layer after the encoder.
        """
        example_inputter = inputters.ExampleInputter(inputter, TagsInputter())
        super().__init__(example_inputter)
        self.encoder = encoder
        self.crf_decoding = crf_decoding
        self.tagging_scheme = None
        self.transition_params = None

    def initialize(self, data_config, params=None):
        self.tagging_scheme = data_config.get("tagging_scheme")
        if self.tagging_scheme:
            self.tagging_scheme = self.tagging_scheme.lower()
        super().initialize(data_config, params=params)

    def build(self, input_shape):
        super().build(input_shape)
        num_tags = self.labels_inputter.vocabulary_size
        self.output_layer = tf.keras.layers.Dense(num_tags)
        if self.crf_decoding:
            self.transition_params = self.add_weight(
                "transition_params", [num_tags, num_tags]
            )

    def call(self, features, labels=None, training=None, step=None):
        length = self.features_inputter.get_length(features)
        inputs = self.features_inputter(features, training=training)
        outputs, _, length = self.encoder(
            inputs, sequence_length=length, training=training
        )
        logits = self.output_layer(outputs)
        if not training:
            if self.crf_decoding:
                tags_id, _ = tfa.text.crf_decode(logits, self.transition_params, length)
                tags_id = tf.cast(tags_id, tf.int64)
            else:
                tags_prob = tf.nn.softmax(logits)
                tags_id = tf.argmax(tags_prob, axis=2)
            predictions = {
                "length": tf.identity(length),
                "tags": self.labels_inputter.ids_to_tokens.lookup(tags_id),
                "tags_id": tags_id,
            }
        else:
            predictions = None
        return logits, predictions

    def compute_loss(self, outputs, labels, training=True):
        if self.crf_decoding:
            log_likelihood, _ = tfa.text.crf_log_likelihood(
                outputs,
                tf.cast(labels["tags_id"], tf.int32),
                labels["length"],
                transition_params=self.transition_params,
            )
            batch_size = tf.shape(log_likelihood)[0]
            return tf.reduce_sum(-log_likelihood) / tf.cast(
                batch_size, log_likelihood.dtype
            )
        else:
            return cross_entropy_sequence_loss(
                outputs,
                labels["tags_id"],
                sequence_length=labels["length"],
                sequence_weight=labels.get("weight"),
                label_smoothing=self.params.get("label_smoothing", 0.0),
                average_in_time=self.params.get("average_loss_in_time", False),
                training=training,
            )

    def get_metrics(self):
        metrics = {"accuracy": tf.keras.metrics.Accuracy()}
        if self.tagging_scheme in ("bioes",):
            f1 = F1()
            metrics["f1"] = f1
            metrics["precision"] = f1.precision
            metrics["recall"] = f1.recall
        return metrics

    def update_metrics(self, metrics, predictions, labels):
        weights = tf.sequence_mask(
            labels["length"], maxlen=tf.shape(labels["tags"])[1], dtype=tf.float32
        )

        metrics["accuracy"].update_state(
            labels["tags_id"], predictions["tags_id"], sample_weight=weights
        )

        if self.tagging_scheme in ("bioes",):
            flag_fn = None
            if self.tagging_scheme == "bioes":
                flag_fn = flag_bioes_tags

            gold_flags, predicted_flags = tf.numpy_function(
                flag_fn,
                [labels["tags"], predictions["tags"], labels["length"]],
                [tf.bool, tf.bool],
            )

            metrics["f1"].update_state(gold_flags, predicted_flags)

    def format_prediction(self, prediction, params=None):
        tags = prediction["tags"][: prediction["length"]]
        return " ".join(tag.decode("utf-8") for tag in tags)


class TagsInputter(inputters.TextInputter):
    """Reading space-separated tags."""

    def __init__(self):
        super().__init__(num_oov_buckets=0)

    def make_features(self, element=None, features=None, training=None):
        features = super().make_features(
            element=element, features=features, training=training
        )
        return {
            "length": features["length"],
            "tags": features["tokens"],
            "tags_id": self.tokens_to_ids.lookup(features["tokens"]),
        }


class F1(tf.keras.metrics.Metric):
    """Defines a F1 metric."""

    def __init__(self, **kwargs):
        """Initializes the metric.

        Args:
          **kwargs: Base class arguments.
        """
        super().__init__(**kwargs)
        self.precision = tf.keras.metrics.Precision()
        self.recall = tf.keras.metrics.Recall()

    @property
    def updates(self):
        """Metric update operations."""
        return self.precision.updates + self.recall.updates

    def update_state(self, y_true, y_pred):
        """Updates the metric state."""
        self.precision.update_state(y_true, y_pred)
        self.recall.update_state(y_true, y_pred)

    def result(self):
        """Returns the metric result."""
        precision = self.precision.result()
        recall = self.recall.result()
        return (2 * precision * recall) / (recall + precision)


def flag_bioes_tags(gold, predicted, sequence_length=None):
    """Flags chunk matches for the BIOES tagging scheme.

    This function will produce the gold flags and the predicted flags. For each aligned
    gold flag ``g`` and predicted flag ``p``:

    * when ``g == p == True``, the chunk has been correctly identified (true positive).
    * when ``g == False and p == True``, the chunk has been incorrectly identified (false positive).
    * when ``g == True and p == False``, the chunk has been missed (false negative).
    * when ``g == p == False``, the chunk has been correctly ignored (true negative).

    Args:
      gold: The gold tags as a Numpy 2D string array.
      predicted: The predicted tags as a Numpy 2D string array.
      sequence_length: The length of each sequence as Numpy array.

    Returns:
      A tuple ``(gold_flags, predicted_flags)``.
    """
    gold_flags = []
    predicted_flags = []

    def _add_true_positive():
        gold_flags.append(True)
        predicted_flags.append(True)

    def _add_false_positive():
        gold_flags.append(False)
        predicted_flags.append(True)

    def _add_true_negative():
        gold_flags.append(False)
        predicted_flags.append(False)

    def _add_false_negative():
        gold_flags.append(True)
        predicted_flags.append(False)

    def _match(ref, hyp, index, length):
        if ref[index].startswith(b"B"):
            match = True
            while index < length and not ref[index].startswith(b"E"):
                if ref[index] != hyp[index]:
                    match = False
                index += 1
            match = match and index < length and ref[index] == hyp[index]
            return match, index
        return ref[index] == hyp[index], index

    for b in range(gold.shape[0]):
        length = sequence_length[b] if sequence_length is not None else gold.shape[1]

        # First pass to detect true positives and true/false negatives.
        index = 0
        while index < length:
            gold_tag = gold[b][index]
            match, index = _match(gold[b], predicted[b], index, length)
            if match:
                if gold_tag == b"O":
                    _add_true_negative()
                else:
                    _add_true_positive()
            else:
                if gold_tag != b"O":
                    _add_false_negative()
            index += 1

        # Second pass to detect false postives.
        index = 0
        while index < length:
            pred_tag = predicted[b][index]
            match, index = _match(predicted[b], gold[b], index, length)
            if not match and pred_tag != b"O":
                _add_false_positive()
            index += 1

    return np.array(gold_flags), np.array(predicted_flags)



"""Define generic inputters."""

import abc

import tensorflow as tf

from opennmt.data import dataset as dataset_util
from opennmt.layers import common
from opennmt.layers.reducer import ConcatReducer, JoinReducer
from opennmt.utils import misc


class Inputter(tf.keras.layers.Layer):
    """Base class for inputters."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._asset_prefix = ""

    @property
    def asset_prefix(self):
        r"""The asset prefix is used to differentiate resources of parallel inputters.
        The most basic examples are the "source\_" and "target\_" prefixes.

        - When reading the data configuration, the inputter will read fields that
          start with this prefix (e.g. "source_vocabulary").
        - Assets exported by this inputter start with this prefix.
        """
        return self._asset_prefix

    @asset_prefix.setter
    def asset_prefix(self, asset_prefix):
        """Sets the asset prefix for this inputter."""
        self._asset_prefix = asset_prefix

    @property
    def num_outputs(self):
        """The number of parallel outputs produced by this inputter."""
        return 1

    def initialize(self, data_config):
        """Initializes the inputter.

        Args:
          data_config: A dictionary containing the data configuration set
            by the user.
        """
        _ = data_config
        return

    def export_assets(self, asset_dir):
        """Exports assets used by this tokenizer.

        Args:
          asset_dir: The directory where assets can be written.

        Returns:
          A dictionary containing additional assets used by the inputter.
        """
        _ = asset_dir
        return {}

    @abc.abstractmethod
    def make_dataset(self, data_file, training=None):
        """Creates the base dataset required by this inputter.

        Args:
          data_file: The data file.
          training: Run in training mode.

        Returns:
          A ``tf.data.Dataset`` instance or a list of ``tf.data.Dataset`` instances.
        """
        raise NotImplementedError()

    def get_dataset_size(self, data_file):
        """Returns the dataset size.

        If the inputter can efficiently compute the dataset size from a training
        file on disk, it can optionally override this method. Otherwise, we may
        compute the size later with a generic and slower approach (iterating over
        the dataset instance).

        Args:
          data_file: The data file.

        Returns:
          The dataset size or ``None``.
        """
        _ = data_file
        return None

    def make_inference_dataset(
        self,
        features_file,
        batch_size,
        batch_type="examples",
        length_bucket_width=None,
        num_threads=1,
        prefetch_buffer_size=None,
    ):
        """Builds a dataset to be used for inference.

        For evaluation and training datasets, see
        :class:`opennmt.inputters.ExampleInputter`.

        Args:
          features_file: The test file.
          batch_size: The batch size to use.
          batch_type: The batching strategy to use: can be "examples" or "tokens".
          length_bucket_width: The width of the length buckets to select batch
            candidates from (for efficiency). Set ``None`` to not constrain batch
            formation.
          num_threads: The number of elements processed in parallel.
          prefetch_buffer_size: The number of batches to prefetch asynchronously. If
            ``None``, use an automatically tuned value.

        Returns:
          A ``tf.data.Dataset``.

        See Also:
          :func:`opennmt.data.inference_pipeline`
        """
        transform_fns = _get_dataset_transforms(
            self, num_threads=num_threads, training=False
        )
        dataset = self.make_dataset(features_file, training=False)
        dataset = dataset.apply(
            dataset_util.inference_pipeline(
                batch_size,
                batch_type=batch_type,
                transform_fns=transform_fns,
                length_bucket_width=length_bucket_width,
                length_fn=self.get_length,
                num_threads=num_threads,
                prefetch_buffer_size=prefetch_buffer_size,
            )
        )
        return dataset

    @abc.abstractmethod
    def input_signature(self):
        """Returns the input signature of this inputter."""
        raise NotImplementedError()

    def get_length(self, features, ignore_special_tokens=False):
        """Returns the length of the input features, if defined.

        Args:
          features: The dictionary of input features.
          ignore_special_tokens: Ignore special tokens that were added by the
            inputter (e.g. <s> and/or </s>).

        Returns:
          The length.
        """
        _ = ignore_special_tokens
        return features.get("length")

    def get_padded_shapes(self, element_spec, maximum_length=None):
        """Returns the padded shapes for dataset elements.

        For example, this is used during batch size autotuning to pad all batches
        to the maximum sequence length.

        Args:
          element_spec: A nested structure of ``tf.TensorSpec``.
          maximum_length: Pad batches to this maximum length.

        Returns:
          A nested structure of ``tf.TensorShape``.
        """
        return tf.nest.map_structure(
            lambda spec: spec.shape
            if spec.shape.rank == 0
            else tf.TensorShape([maximum_length]).concatenate(spec.shape[1:]),
            element_spec,
        )

    def has_prepare_step(self):
        """Returns ``True`` if this inputter implements a data preparation step
        in method :meth:`opennmt.inputters.Inputter.prepare_elements`.
        """
        return False

    def prepare_elements(self, elements, training=None):
        """Prepares dataset elements.

        This method is called on a batch of dataset elements. For example, it
        can be overriden to apply an external pre-tokenization.

        Note that the results of the method are unbatched and then passed to
        method :meth:`opennmt.inputters.Inputter.make_features`.

        Args:
          elements: A batch of dataset elements.
          training: Run in training mode.

        Returns:
          A (possibly nested) structure of ``tf.Tensor``.
        """
        return elements

    @abc.abstractmethod
    def make_features(self, element=None, features=None, training=None):
        """Creates features from data.

        This is typically called in a data pipeline (such as ``Dataset.map``).
        Common transformation includes tokenization, parsing, vocabulary lookup,
        etc.

        This method accepts both a single :obj:`element` from the dataset or a
        partially built dictionary of :obj:`features`.

        Args:
          element: An element from the dataset returned by
            :meth:`opennmt.inputters.Inputter.make_dataset`.
          features: An optional and possibly partial dictionary of features to
            augment.
          training: Run in training mode.

        Returns:
          A dictionary of ``tf.Tensor``.
        """
        raise NotImplementedError()

    def keep_for_training(self, features, maximum_length=None):
        """Returns ``True`` if this example should be kept for training.

        Args:
          features: A dictionary of ``tf.Tensor``.
          maximum_length: The maximum length used for training.

        Returns:
          A boolean.
        """
        length = self.get_length(features)
        if length is None:
            return True
        is_valid = tf.greater(length, 0)
        if maximum_length is not None:
            is_valid = tf.logical_and(is_valid, tf.less_equal(length, maximum_length))
        return is_valid

    def call(self, features, training=None):
        """Creates the model input from the features (e.g. word embeddings).

        Args:
          features: A dictionary of ``tf.Tensor``, the output of
            :meth:`opennmt.inputters.Inputter.make_features`.
          training: Run in training mode.

        Returns:
          The model input.
        """
        _ = training
        return features

    def visualize(self, model_root, log_dir):
        """Visualizes the transformation, usually embeddings.

        Args:
          model_root: The root model object.
          log_dir: The active log directory.
        """
        _ = model_root
        _ = log_dir
        return


class MultiInputter(Inputter):
    """An inputter that gathers multiple inputters, possibly nested."""

    def __init__(self, inputters, reducer=None):
        if not isinstance(inputters, list) or not inputters:
            raise ValueError("inputters must be a non empty list")
        dtype = inputters[0].dtype
        for inputter in inputters:
            if inputter.dtype != dtype:
                raise TypeError("All inputters must have the same dtype")
        super().__init__(dtype=dtype)
        self.inputters = inputters
        self.reducer = reducer
        self.asset_prefix = ""  # Generate the default prefix for sub-inputters.

    @Inputter.asset_prefix.setter
    def asset_prefix(self, asset_prefix):
        self._asset_prefix = asset_prefix
        for i, inputter in enumerate(self.inputters):
            inputter.asset_prefix = "%s%d_" % (asset_prefix, i + 1)

    @property
    def num_outputs(self):
        if self.reducer is None or isinstance(self.reducer, JoinReducer):
            return len(self.inputters)
        return 1

    def get_leaf_inputters(self):
        """Returns a list of all leaf Inputter instances."""
        inputters = []
        for inputter in self.inputters:
            if isinstance(inputter, MultiInputter):
                inputters.extend(inputter.get_leaf_inputters())
            else:
                inputters.append(inputter)
        return inputters

    def __getattribute__(self, name):
        if name == "built":
            return all(inputter.built for inputter in self.inputters)
        else:
            return super().__getattribute__(name)

    def initialize(self, data_config):
        for inputter in self.inputters:
            inputter.initialize(
                misc.RelativeConfig(
                    data_config, inputter.asset_prefix, config_name="data"
                )
            )

    def export_assets(self, asset_dir):
        assets = {}
        for inputter in self.inputters:
            assets.update(inputter.export_assets(asset_dir))
        return assets

    def has_prepare_step(self):
        return any(inputter.has_prepare_step() for inputter in self.inputters)

    def prepare_elements(self, elements, training=None):
        return tuple(
            inputter.prepare_elements(elts)
            for inputter, elts in zip(self.inputters, elements)
        )

    def visualize(self, model_root, log_dir):
        for inputter in self.inputters:
            inputter.visualize(model_root, log_dir)


class ParallelInputter(MultiInputter):
    """A multi inputter that processes parallel data."""

    def __init__(
        self, inputters, reducer=None, share_parameters=False, combine_features=True
    ):
        """Initializes a parallel inputter.

        Args:
          inputters: A list of :class:`opennmt.inputters.Inputter`.
          reducer: A :class:`opennmt.layers.Reducer` to merge all inputs. If
            set, parallel inputs are assumed to have the same length.
          share_parameters: Share the inputters parameters.
          combine_features: Combine each inputter features in a single dict or
            return them separately. This is typically ``True`` for multi source
            inputs but ``False`` for features/labels parallel data.

        Raises:
          ValueError: if :obj:`share_parameters` is set but the child inputters are
            not of the same type.
        """
        super().__init__(inputters, reducer=reducer)
        self.combine_features = combine_features
        self.share_parameters = share_parameters
        if self.share_parameters:
            leaves = self.get_leaf_inputters()
            for inputter in leaves[1:]:
                if type(inputter) is not type(leaves[0]):  # noqa: E721
                    raise ValueError(
                        "Each inputter must be of the same type for parameter sharing"
                    )

    def _structure(self):
        """Returns the nested structure that represents this parallel inputter."""
        return [
            inputter._structure() if isinstance(inputter, ParallelInputter) else None
            for inputter in self.inputters
        ]

    def make_dataset(self, data_file, training=None):
        if not isinstance(data_file, list):
            data_file = [data_file]

        # For evaluation and inference, accept a flat list of data files for nested inputters.
        # This is needed when nesting can't easily be represented (e.g. on the command line).
        if not training:
            try:
                data_file = tf.nest.pack_sequence_as(
                    self._structure(), tf.nest.flatten(data_file)
                )
            except ValueError:
                data_file = []  # This will raise the error below.

        if len(data_file) != len(self.inputters):
            raise ValueError(
                "The number of data files must be the same as the number of inputters"
            )

        num_files = -1
        datasets = []
        for i, (inputter, data) in enumerate(zip(self.inputters, data_file)):
            dataset = inputter.make_dataset(data, training=training)
            if not isinstance(dataset, list):
                dataset = [dataset]
            datasets.append(dataset)
            if num_files < 0:
                num_files = len(dataset)
            elif len(dataset) != num_files:
                raise ValueError(
                    "All parallel inputs must have the same number of data files, "
                    "saw %d files for input 0 but got %d files for input %d"
                    % (num_files, len(dataset), i)
                )

        parallel_datasets = [
            tf.data.Dataset.zip(tuple(parallel_dataset))
            for parallel_dataset in zip(*datasets)
        ]
        if len(parallel_datasets) == 1:
            return parallel_datasets[0]
        if not training:
            raise ValueError("Only training data can be configured to multiple files")
        return parallel_datasets

    def get_dataset_size(self, data_file):
        common_size = None
        for inputter, data in zip(self.inputters, data_file):
            size = inputter.get_dataset_size(data)
            if size is not None:
                if common_size is None:
                    common_size = size
                elif size != common_size:
                    raise RuntimeError("Parallel datasets do not have the same size")
        return common_size

    def input_signature(self):
        if self.combine_features:
            signature = {}
            for i, inputter in enumerate(self.inputters):
                for key, value in inputter.input_signature().items():
                    signature["{}_{}".format(key, i)] = value
            return signature
        else:
            return tuple(inputter.input_signature() for inputter in self.inputters)

    def _index_features(self, features, index):
        if self.combine_features:
            return misc.extract_prefixed_keys(features, "inputter_{}_".format(index))
        else:
            return features[index]

    def get_length(self, features, ignore_special_tokens=False):
        lengths = [
            inputter.get_length(
                self._index_features(features, i),
                ignore_special_tokens=ignore_special_tokens,
            )
            for i, inputter in enumerate(self.inputters)
        ]
        if self.reducer is None:
            return lengths
        else:
            return lengths[0]

    def get_padded_shapes(self, element_spec, maximum_length=None):
        if maximum_length is None:
            maximum_length = [None for _ in self.inputters]
        elif not isinstance(maximum_length, (list, tuple)) or len(
            maximum_length
        ) != len(self.inputters):
            raise ValueError(
                "A maximum length should be set for each parallel inputter"
            )
        if self.combine_features:
            shapes = {}
            for i, (inputter, length) in enumerate(zip(self.inputters, maximum_length)):
                prefix = "inputter_%d_" % i
                spec = misc.extract_prefixed_keys(element_spec, prefix)
                sub_shapes = inputter.get_padded_shapes(spec, maximum_length=length)
                for key, value in sub_shapes.items():
                    shapes["%s%s" % (prefix, key)] = value
            return shapes
        else:
            return type(element_spec)(
                inputter.get_padded_shapes(spec, maximum_length=length)
                for inputter, spec, length in zip(
                    self.inputters, element_spec, maximum_length
                )
            )

    def make_features(self, element=None, features=None, training=None):
        if self.combine_features:
            if features is None:
                features = {}
            for i, inputter in enumerate(self.inputters):
                prefix = "inputter_%d_" % i
                sub_features = misc.extract_prefixed_keys(features, prefix)
                if not sub_features:
                    # Also try to read the format produced by the serving features.
                    sub_features = misc.extract_suffixed_keys(features, "_%d" % i)
                sub_features = inputter.make_features(
                    element=element[i] if element is not None else None,
                    features=sub_features,
                    training=training,
                )
                for key, value in sub_features.items():
                    features["%s%s" % (prefix, key)] = value
            return features
        else:
            if features is None:
                features = [{} for _ in self.inputters]
            else:
                features = list(features)
            for i, inputter in enumerate(self.inputters):
                features[i] = inputter.make_features(
                    element=element[i] if element is not None else None,
                    features=features[i],
                    training=training,
                )
            return tuple(features)

    def keep_for_training(self, features, maximum_length=None):
        if not isinstance(maximum_length, list):
            maximum_length = [maximum_length]
        # Unset maximum lengths are set to None (i.e. no constraint).
        maximum_length += [None] * (len(self.inputters) - len(maximum_length))
        constraints = []
        for i, inputter in enumerate(self.inputters):
            keep = inputter.keep_for_training(
                self._index_features(features, i), maximum_length=maximum_length[i]
            )
            if isinstance(keep, bool):
                if not keep:
                    return False
                continue
            constraints.append(keep)
        if not constraints:
            return True
        return tf.reduce_all(constraints)

    def build(self, input_shape):
        if self.share_parameters:
            # When sharing parameters, build the first leaf inputter and then set
            # all attributes with parameters to the other inputters.
            leaves = self.get_leaf_inputters()
            first, others = leaves[0], leaves[1:]
            first.build(input_shape)
            for name, attr in first.__dict__.copy().items():
                if isinstance(attr, tf.Variable) or (
                    isinstance(attr, tf.Module) and attr.variables
                ):
                    for inputter in others:
                        setattr(inputter, name, attr)
                        inputter.built = True
        else:
            for inputter in self.inputters:
                inputter.build(input_shape)
        super().build(input_shape)

    def call(self, features, training=None):
        transformed = [
            inputter(self._index_features(features, i), training=training)
            for i, inputter in enumerate(self.inputters)
        ]
        if self.reducer is not None:
            transformed = self.reducer(transformed)
        return transformed


class MixedInputter(MultiInputter):
    """An multi inputter that applies several transformation on the same data
    (e.g. combine word-level and character-level embeddings).
    """

    def __init__(self, inputters, reducer=ConcatReducer(), dropout=0.0):
        """Initializes a mixed inputter.

        Args:
          inputters: A list of :class:`opennmt.inputters.Inputter`.
          reducer: A :class:`opennmt.layers.Reducer` to merge all inputs.
          dropout: The probability to drop units in the merged inputs.
        """
        super().__init__(inputters, reducer=reducer)
        self.dropout = dropout

    def make_dataset(self, data_file, training=None):
        datasets = [
            inputter.make_dataset(data_file, training=training)
            for inputter in self.inputters
        ]
        for dataset in datasets[1:]:
            if not isinstance(dataset, datasets[0].__class__):
                raise ValueError(
                    "All inputters should use the same dataset in a MixedInputter setting"
                )
        return datasets[0]

    def get_dataset_size(self, data_file):
        for inputter in self.inputters:
            size = inputter.get_dataset_size(data_file)
            if size is not None:
                return size
        return None

    def input_signature(self):
        signature = {}
        for inputter in self.inputters:
            signature.update(inputter.input_signature())
        return signature

    def get_length(self, features, ignore_special_tokens=False):
        return self.inputters[0].get_length(
            features, ignore_special_tokens=ignore_special_tokens
        )

    def make_features(self, element=None, features=None, training=None):
        if features is None:
            features = {}
        for inputter in self.inputters:
            features = inputter.make_features(
                element=element, features=features, training=training
            )
        return features

    def build(self, input_shape):
        for inputter in self.inputters:
            inputter.build(input_shape)
        super().build(input_shape)

    def call(self, features, training=None):
        transformed = []
        for inputter in self.inputters:
            transformed.append(inputter(features, training=training))
        outputs = self.reducer(transformed)
        outputs = common.dropout(outputs, self.dropout, training=training)
        return outputs


class ExampleInputterAdapter:
    """Extends an inputter with methods to build evaluation and training datasets."""

    def make_evaluation_dataset(
        self,
        features_file,
        labels_file,
        batch_size,
        batch_type="examples",
        length_bucket_width=None,
        num_threads=1,
        prefetch_buffer_size=None,
    ):
        """Builds a dataset to be used for evaluation.

        Args:
          features_file: The evaluation source file.
          labels_file: The evaluation target file.
          batch_size: The batch size to use.
          batch_type: The batching strategy to use: can be "examples" or "tokens".
          length_bucket_width: The width of the length buckets to select batch
            candidates from (for efficiency). Set ``None`` to not constrain batch
            formation.
          num_threads: The number of elements processed in parallel.
          prefetch_buffer_size: The number of batches to prefetch asynchronously. If
            ``None``, use an automatically tuned value.

        Returns:
          A ``tf.data.Dataset``.

        See Also:
          :func:`opennmt.data.inference_pipeline`
        """
        if labels_file is not None:
            data_files = [features_file, labels_file]
            length_fn = [
                self.features_inputter.get_length,
                self.labels_inputter.get_length,
            ]
        else:
            data_files = features_file
            length_fn = self.get_length

        transform_fns = _get_dataset_transforms(
            self, num_threads=num_threads, training=False
        )
        dataset = self.make_dataset(data_files, training=False)
        dataset = dataset.apply(
            dataset_util.inference_pipeline(
                batch_size,
                batch_type=batch_type,
                transform_fns=transform_fns,
                length_bucket_width=length_bucket_width,
                length_fn=length_fn,
                num_threads=num_threads,
                prefetch_buffer_size=prefetch_buffer_size,
            )
        )
        return dataset

    def make_training_dataset(
        self,
        features_file,
        labels_file,
        batch_size,
        batch_type="examples",
        batch_multiplier=1,
        batch_size_multiple=1,
        shuffle_buffer_size=None,
        length_bucket_width=None,
        maximum_features_length=None,
        maximum_labels_length=None,
        single_pass=False,
        num_shards=1,
        shard_index=0,
        num_threads=4,
        prefetch_buffer_size=None,
        cardinality_multiple=1,
        weights=None,
        batch_autotune_mode=False,
    ):
        """Builds a dataset to be used for training. It supports the full training
        pipeline, including:

        * sharding
        * shuffling
        * filtering
        * bucketing
        * prefetching

        Args:
          features_file: The source file or a list of training source files.
          labels_file: The target file or a list of training target files.
          batch_size: The batch size to use.
          batch_type: The training batching strategy to use: can be "examples" or
            "tokens".
          batch_multiplier: The batch size multiplier to prepare splitting accross
             replicated graph parts.
          batch_size_multiple: When :obj:`batch_type` is "tokens", ensure that the
            resulting batch size is a multiple of this value.
          shuffle_buffer_size: The number of elements from which to sample.
          length_bucket_width: The width of the length buckets to select batch
            candidates from (for efficiency). Set ``None`` to not constrain batch
            formation.
          maximum_features_length: The maximum length or list of maximum lengths of
            the features sequence(s). ``None`` to not constrain the length.
          maximum_labels_length: The maximum length of the labels sequence.
            ``None`` to not constrain the length.
          single_pass: If ``True``, makes a single pass over the training data.
          num_shards: The number of data shards (usually the number of workers in a
            distributed setting).
          shard_index: The shard index this data pipeline should read from.
          num_threads: The number of elements processed in parallel.
          prefetch_buffer_size: The number of batches to prefetch asynchronously. If
            ``None``, use an automatically tuned value.
          cardinality_multiple: Ensure that the dataset cardinality is a multiple of
            this value when :obj:`single_pass` is ``True``.
          weights: An optional list of weights to create a weighted dataset out of
            multiple training files.
          batch_autotune_mode: When enabled, all batches are padded to the maximum
            sequence length.

        Returns:
          A ``tf.data.Dataset``.

        See Also:
          :func:`opennmt.data.training_pipeline`
        """
        if labels_file is not None:
            data_files = [features_file, labels_file]
            maximum_length = [maximum_features_length, maximum_labels_length]
            features_length_fn = self.features_inputter.get_length
            labels_length_fn = self.labels_inputter.get_length
        else:
            data_files = features_file
            maximum_length = maximum_features_length
            features_length_fn = self.get_length
            labels_length_fn = None

        dataset = self.make_dataset(data_files, training=True)

        filter_fn = lambda *arg: (
            self.keep_for_training(
                misc.item_or_tuple(arg), maximum_length=maximum_length
            )
        )

        transform_fns = _get_dataset_transforms(
            self, num_threads=num_threads, training=True
        )
        transform_fns.append(lambda dataset: dataset.filter(filter_fn))

        if batch_autotune_mode:
            # In this mode we want to return batches where all sequences are padded
            # to the maximum possible length in order to maximize the memory usage.
            # Shuffling, sharding, prefetching, etc. are not applied since correctness and
            # performance are not important.

            if isinstance(dataset, list):  # Ignore weighted dataset.
                dataset = dataset[0]

            # We repeat the dataset now to ensure full batches are always returned.
            dataset = dataset.repeat()
            for transform_fn in transform_fns:
                dataset = dataset.apply(transform_fn)

            # length_fn returns the maximum length instead of the actual example length so
            # that batches are built as if each example has the maximum length.
            if labels_file is not None:
                constant_length_fn = [
                    lambda x: maximum_features_length,
                    lambda x: maximum_labels_length,
                ]
            else:
                constant_length_fn = lambda x: maximum_features_length

            # The length dimension is set to the maximum length in the padded shapes.
            padded_shapes = self.get_padded_shapes(
                dataset.element_spec, maximum_length=maximum_length
            )
            dataset = dataset.apply(
                dataset_util.batch_sequence_dataset(
                    batch_size,
                    batch_type=batch_type,
                    batch_multiplier=batch_multiplier,
                    length_bucket_width=1,
                    length_fn=constant_length_fn,
                    padded_shapes=padded_shapes,
                )
            )
            return dataset

        if weights is not None:
            dataset = (dataset, weights)
        dataset = dataset_util.training_pipeline(
            batch_size,
            batch_type=batch_type,
            batch_multiplier=batch_multiplier,
            batch_size_multiple=batch_size_multiple,
            transform_fns=transform_fns,
            length_bucket_width=length_bucket_width,
            features_length_fn=features_length_fn,
            labels_length_fn=labels_length_fn,
            single_pass=single_pass,
            num_shards=num_shards,
            shard_index=shard_index,
            num_threads=num_threads,
            dataset_size=self.get_dataset_size(data_files),
            shuffle_buffer_size=shuffle_buffer_size,
            prefetch_buffer_size=prefetch_buffer_size,
            cardinality_multiple=cardinality_multiple,
        )(dataset)
        return dataset


def _register_example_weight(features, labels, weight):
    labels["weight"] = tf.strings.to_number(weight)
    return features, labels


class ExampleInputter(ParallelInputter, ExampleInputterAdapter):
    """An inputter that returns training examples (parallel features and labels)."""

    def __init__(
        self,
        features_inputter,
        labels_inputter,
        share_parameters=False,
        accepted_annotations=None,
    ):
        """Initializes this inputter.

        Args:
          features_inputter: An inputter producing the features (source).
          labels_inputter: An inputter producing the labels (target).
          share_parameters: Share the inputters parameters.
          accepted_annotations: An optional dictionary mapping annotation names in
            the data configuration (e.g. "train_alignments") to a callable with
            signature ``(features, labels, annotations) -> (features, labels)``.
        """
        self.features_inputter = features_inputter
        self.labels_inputter = labels_inputter
        super().__init__(
            [self.features_inputter, self.labels_inputter],
            share_parameters=share_parameters,
            combine_features=False,
        )
        # Set a meaningful prefix for source and target.
        self.features_inputter.asset_prefix = "source_"
        self.labels_inputter.asset_prefix = "target_"

        self.accepted_annotations = accepted_annotations or {}
        self.accepted_annotations["example_weights"] = _register_example_weight
        self.annotation_files = {}

    def initialize(self, data_config):
        super().initialize(data_config)

        # Check if some accepted annotations are defined in the data configuration.
        for annotation in self.accepted_annotations.keys():
            path = data_config.get(annotation)
            if path is not None:
                self.annotation_files[annotation] = path

    def make_dataset(self, data_file, training=None):
        dataset = super().make_dataset(data_file, training=training)
        if not training or not self.annotation_files:
            return dataset

        # Some annotations are configured and should be zipped to the training dataset.
        all_annotation_datasets = tf.nest.map_structure(
            tf.data.TextLineDataset, self.annotation_files
        )

        # Common case of a non-weighted dataset.
        if not isinstance(dataset, list):
            return tf.data.Dataset.zip({"examples": dataset, **all_annotation_datasets})

        # Otherwise, there should be as many annotations datasets as input datasets.
        datasets = dataset
        for name, annotation_datasets in all_annotation_datasets.items():
            num_annotation_datasets = (
                len(annotation_datasets) if isinstance(annotation_datasets, list) else 1
            )
            if num_annotation_datasets != len(datasets):
                raise ValueError(
                    "%d '%s' files were provided, but %d were expected to match the "
                    "number of data files"
                    % (num_annotation_datasets, name, len(datasets))
                )

        # Convert dict of lists to list of dicts.
        all_annotation_datasets = [
            dict(zip(all_annotation_datasets, t))
            for t in zip(*all_annotation_datasets.values())
        ]

        return [
            tf.data.Dataset.zip({"examples": dataset, **annotation_datasets})
            for dataset, annotation_datasets in zip(datasets, all_annotation_datasets)
        ]

    def get_dataset_size(self, data_file):
        size = super().get_dataset_size(data_file)
        if size is not None:
            for annotation, path in self.annotation_files.items():
                annotation_size = tf.nest.map_structure(misc.count_lines, path)
                if size != annotation_size:
                    raise RuntimeError(
                        "Annotation dataset '%s' does not have the same size as "
                        "the examples dataset" % annotation
                    )
        return size

    def make_features(self, element=None, features=None, training=None):
        if training and self.annotation_files:
            annotations = element.copy()
            example = annotations.pop("examples")
        else:
            annotations = {}
            example = element

        features, labels = super().make_features(
            element=example, features=features, training=training
        )

        # Load each annotation into the features and labels dict.
        for name, annotation in annotations.items():
            features, labels = self.accepted_annotations[name](
                features, labels, annotation
            )
        return features, labels

    def make_inference_dataset(
        self,
        features_file,
        batch_size,
        batch_type="examples",
        length_bucket_width=None,
        num_threads=1,
        prefetch_buffer_size=None,
    ):
        return self.features_inputter.make_inference_dataset(
            features_file,
            batch_size,
            batch_type=batch_type,
            length_bucket_width=length_bucket_width,
            num_threads=num_threads,
            prefetch_buffer_size=prefetch_buffer_size,
        )


def _get_dataset_transforms(
    inputter,
    num_threads=None,
    training=None,
    prepare_batch_size=128,
):
    transform_fns = []

    if inputter.has_prepare_step():
        prepare_fn = lambda *arg: inputter.prepare_elements(
            misc.item_or_tuple(arg), training=training
        )
        transform_fns.extend(
            [
                lambda dataset: dataset.batch(prepare_batch_size),
                lambda dataset: dataset.map(prepare_fn, num_parallel_calls=num_threads),
                lambda dataset: dataset.unbatch(),
            ]
        )

    map_fn = lambda *arg: inputter.make_features(
        element=misc.item_or_tuple(arg), training=training
    )
    transform_fns.append(
        lambda dataset: dataset.map(map_fn, num_parallel_calls=num_threads)
    )
    return transform_fns



"""Script that merges configurations for debug or simplification."""

import argparse

import yaml

from opennmt.config import load_config


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("config", nargs="+", help="Configuration files.")
    args = parser.parse_args()
    config = load_config(args.config)
    print(yaml.dump(config, default_flow_style=False))


if __name__ == "__main__":
    main()



"""Define losses."""

import tensorflow as tf


def _smooth_one_hot_labels(logits, labels, label_smoothing):
    num_classes = logits.shape[-1]
    on_value = 1.0 - label_smoothing
    off_value = label_smoothing / (num_classes - 1)
    return tf.one_hot(
        labels,
        num_classes,
        on_value=tf.cast(on_value, logits.dtype),
        off_value=tf.cast(off_value, logits.dtype),
    )


def _softmax_cross_entropy(logits, labels, label_smoothing, training):
    # Computes the softmax in full precision.
    logits = tf.cast(logits, tf.float32)
    if training and label_smoothing > 0.0:
        smoothed_labels = _smooth_one_hot_labels(logits, labels, label_smoothing)
        return tf.nn.softmax_cross_entropy_with_logits(smoothed_labels, logits)
    else:
        return tf.nn.sparse_softmax_cross_entropy_with_logits(labels, logits)


def cross_entropy_sequence_loss(
    logits,
    labels,
    sequence_length=None,
    label_smoothing=0.0,
    average_in_time=False,
    training=None,
    sequence_weight=None,
    mask_outliers=False,
):
    """Computes the cross entropy loss of sequences.

    Args:
      logits: The unscaled probabilities with shape :math:`[B, T, V]`.
      labels: The true labels with shape :math:`[B, T]`.
      sequence_length: The length of each sequence with shape :math:`[B]`.
      label_smoothing: The label smoothing value.
      average_in_time: If ``True``, also average the loss in the time dimension.
      training: Compute training loss.
      sequence_weight: The weight of each sequence with shape :math:`[B]`.
      mask_outliers: Mask large training loss values considered as outliers.

    Returns:
      A tuple (cumulated loss, loss normalizer, token-level normalizer).
    """
    cross_entropy = _softmax_cross_entropy(logits, labels, label_smoothing, training)
    dtype = cross_entropy.dtype

    shape = tf.shape(logits)
    batch_size = shape[0]
    max_time = shape[1]

    if sequence_length is None:
        sequence_length = tf.fill([batch_size], max_time)

    weight = tf.sequence_mask(sequence_length, maxlen=max_time, dtype=dtype)

    if training and mask_outliers:
        import tensorflow_probability as tfp

        # Outliers are detected using the interquantile range (IQR).
        examples_loss = tf.reduce_sum(cross_entropy * weight, axis=-1)
        examples_score = examples_loss / tf.reduce_sum(weight, axis=-1)
        percentiles = tfp.stats.percentile(examples_score, [25, 75])
        iqr = percentiles[1] - percentiles[0]
        threshold = percentiles[1] + 1.5 * iqr
        if sequence_weight is None:
            sequence_weight = tf.ones([batch_size], dtype=dtype)
        sequence_weight = tf.where(
            examples_score > threshold,
            x=tf.zeros_like(sequence_weight),
            y=sequence_weight,
        )

    if sequence_weight is not None:
        sequence_weight = tf.cast(sequence_weight, dtype)
        weight *= tf.expand_dims(sequence_weight, 1)

    loss = tf.reduce_sum(cross_entropy * weight)

    loss_token_normalizer = tf.reduce_sum(weight)
    if average_in_time or not training:
        loss_normalizer = loss_token_normalizer
    elif sequence_weight is not None:
        loss_normalizer = tf.reduce_sum(sequence_weight)
    else:
        loss_normalizer = tf.cast(batch_size, dtype)

    return loss, loss_normalizer, loss_token_normalizer


def cross_entropy_loss(logits, labels, label_smoothing=0.0, training=None, weight=None):
    """Computes the cross entropy loss.

    Args:
      logits: The unscaled probabilities with shape :math:`[B, V]`.
      labels: The true labels with shape :math:`[B]`.
      label_smoothing: The label smoothing value.
      training: Compute training loss.
      weight: The weight of each example with shape :math:`[B]`.

    Returns:
      The cumulated loss and the loss normalizer.
    """
    cross_entropy = _softmax_cross_entropy(logits, labels, label_smoothing, training)

    if weight is not None:
        weight = tf.cast(weight, cross_entropy.dtype)
        cross_entropy *= weight
        loss_normalizer = tf.reduce_sum(weight)
    else:
        batch_size = tf.shape(cross_entropy)[0]
        loss_normalizer = tf.cast(batch_size, cross_entropy.dtype)

    loss = tf.reduce_sum(cross_entropy)
    return loss, loss_normalizer


def guided_alignment_cost(
    attention_probs, gold_alignment, sequence_length=None, cost_type="ce", weight=1
):
    """Computes the guided alignment cost.

    Args:
      attention_probs: The attention probabilities, a float ``tf.Tensor`` of shape
        :math:`[B, T_t, T_s]`.
      gold_alignment: The true alignment matrix, a float ``tf.Tensor`` of shape
        :math:`[B, T_t, T_s]`.
      sequence_length: The length of each sequence.
      cost_type: The type of the cost function to compute (can be: ce, mse).
      weight: The weight applied to the cost.

    Returns:
      The guided alignment cost.

    Raises:
      ValueError: if :obj:`cost_type` is invalid.
    """
    if cost_type == "ce":
        loss = tf.keras.losses.CategoricalCrossentropy(
            reduction=tf.keras.losses.Reduction.SUM
        )
    elif cost_type == "mse":
        loss = tf.keras.losses.MeanSquaredError(reduction=tf.keras.losses.Reduction.SUM)
    else:
        raise ValueError("invalid guided alignment cost: %s" % cost_type)

    if sequence_length is not None:
        sample_weight = tf.sequence_mask(
            sequence_length,
            maxlen=tf.shape(attention_probs)[1],
            dtype=attention_probs.dtype,
        )
        sample_weight = tf.expand_dims(sample_weight, -1)
        normalizer = tf.reduce_sum(sequence_length)
    else:
        sample_weight = None
        normalizer = tf.size(attention_probs)

    cost = loss(gold_alignment, attention_probs, sample_weight=sample_weight)
    cost /= tf.cast(normalizer, cost.dtype)
    return weight * cost


def regularization_penalty(regularization_type, scale, weights):
    """Computes the weights regularization penalty.

    Args:
      regularization_type: The regularization type: ``l1``, ``l2``, or ``l1_l2``.
      scale: The regularization multiplier. If :obj:`regularization_type` is
        ``l1_l2``, this should be a list or tuple containing the L1 regularization
        scale and the L2 regularization scale.
      weights: The list of weights.

    Returns:
      The regularization penalty.

    Raises:
      ValueError: if :obj:`regularization_type` is invalid or is ``l1_l2`` but
        :obj:`scale` is not a sequence.
    """
    regularization_type = regularization_type.lower()
    if regularization_type == "l1":
        regularizer = tf.keras.regularizers.l1(l=float(scale))
    elif regularization_type == "l2":
        regularizer = tf.keras.regularizers.l2(l=float(scale))
    elif regularization_type == "l1_l2":
        if not isinstance(scale, (list, tuple)) or len(scale) != 2:
            raise ValueError("l1_l2 regularization requires 2 scale values")
        regularizer = tf.keras.regularizers.l1_l2(
            l1=float(scale[0]), l2=float(scale[1])
        )
    else:
        raise ValueError("invalid regularization type %s" % regularization_type)

    weights = list(filter(lambda v: not _is_bias(v), weights))
    penalty = tf.add_n([regularizer(w) for w in weights])
    return penalty


def _is_bias(variable):
    return len(variable.shape) == 1 and variable.name.endswith("bias:0")


def _negative_log_likelihood(logits, labels, sequence_length):
    nll_num, nll_den, _ = cross_entropy_sequence_loss(
        logits, labels, sequence_length, average_in_time=True
    )
    return nll_num / nll_den


def max_margin_loss(
    true_logits,
    true_labels,
    true_sequence_length,
    negative_logits,
    negative_labels,
    negative_sequence_length,
    eta=0.1,
):
    """Computes the max-margin loss described in
    https://www.aclweb.org/anthology/P19-1623.

    Args:
      true_logits: The unscaled probabilities from the true example.
      negative_logits: The unscaled probabilities from the negative example.
      true_labels: The true labels.
      true_sequence_length: The length of each true sequence.
      negative_labels: The negative labels.
      negative_sequence_length: The length of each negative sequence.
      eta: Ensure that the margin is higher than this value.

    Returns:
      The max-margin loss.
    """
    true_nll = _negative_log_likelihood(true_logits, true_labels, true_sequence_length)
    negative_nll = _negative_log_likelihood(
        negative_logits, negative_labels, negative_sequence_length
    )
    margin = true_nll - negative_nll + eta
    return tf.maximum(margin, 0)



"""Module defining tokenizers.

Tokenizers can work on string ``tf.Tensor`` as in-graph transformation.
"""

try:
    import pyonmttok

    from opennmt.tokenizers.opennmt_tokenizer import OpenNMTTokenizer
except ImportError:
    pass

try:
    from opennmt.tokenizers.sentencepiece_tokenizer import SentencePieceTokenizer
except ImportError:
    pass

from opennmt.tokenizers.tokenizer import (
    CharacterTokenizer,
    SpaceTokenizer,
    Tokenizer,
    make_tokenizer,
    register_tokenizer,
)



"""Evaluation related classes and functions."""

import collections
import os
import shutil

import tensorflow as tf

from opennmt.utils import exporters, misc
from opennmt.utils import scorers as scorers_lib

_SUMMARIES_SCOPE = "metrics"


class EarlyStopping(
    collections.namedtuple("EarlyStopping", ("metric", "min_improvement", "steps"))
):
    """Conditions for early stopping."""


class Evaluator(object):
    """Model evaluator."""

    def __init__(
        self,
        model,
        features_file,
        labels_file,
        batch_size,
        batch_type="examples",
        length_bucket_width=None,
        scorers=None,
        save_predictions=False,
        early_stopping=None,
        model_dir=None,
        export_on_best=None,
        exporter=None,
        max_exports_to_keep=5,
    ):
        """Initializes the evaluator.

        Args:
          model: A :class:`opennmt.models.Model` to evaluate.
          features_file: Path to the evaluation features.
          labels_file: Path to the evaluation labels.
          batch_size: The evaluation batch size.
          batch_type: The batching strategy to use: can be "examples" or "tokens".
          length_bucket_width: The width of the length buckets to select batch
            candidates from (for efficiency). Set ``None`` to not constrain batch
            formation.
          scorers: A list of scorers, callables taking the path to the reference and
            the hypothesis and return one or more scores.
          save_predictions: Save evaluation predictions to a file. This is ``True``
            when :obj:`scorers` is set.
          early_stopping: An ``EarlyStopping`` instance.
          model_dir: The active model directory.
          export_on_best: Export a model when this evaluation metric has the
            best value so far.
          exporter: A :class:`opennmt.utils.Exporter` instance to export the model.
            Defaults to :class:`opennmt.utils.SavedModelExporter`.
          max_exports_to_keep: Maximum number of exports to keep. Older exports will
            be garbage collected. Set to ``None`` to keep all exports.

        Raises:
          ValueError: If :obj:`save_predictions` is set but the model is not compatible.
          ValueError: If :obj:`save_predictions` is set but :obj:`model_dir` is ``None``.
          ValueError: If :obj:`export_on_best` is set but :obj:`model_dir` is ``None``.
          ValueError: If the :obj:`early_stopping` configuration is invalid.
        """
        if model_dir is not None:
            export_dir = os.path.join(model_dir, "export")
            eval_dir = os.path.join(model_dir, "eval")
        else:
            if save_predictions:
                raise ValueError(
                    "Saving evaluation predictions requires model_dir to be set"
                )
            if export_on_best is not None:
                raise ValueError("Exporting models requires model_dir to be set")
            export_dir = None
            eval_dir = None

        if scorers is None:
            scorers = []
        if scorers:
            save_predictions = True
        if save_predictions:
            if model.unsupervised:
                raise ValueError(
                    "This model does not support saving evaluation predictions"
                )
            if not tf.io.gfile.exists(eval_dir):
                tf.io.gfile.makedirs(eval_dir)
        self._model = model
        self._labels_file = labels_file
        self._save_predictions = save_predictions
        self._scorers = scorers
        self._eval_dir = eval_dir
        self._metrics_history = []
        if eval_dir is not None:
            self._summary_writer = tf.summary.create_file_writer(eval_dir)
            summaries = misc.read_summaries(eval_dir)
            for step, values in summaries:
                metrics = misc.extract_prefixed_keys(values, _SUMMARIES_SCOPE + "/")
                self._metrics_history.append((step, metrics))
        else:
            self._summary_writer = tf.summary.create_noop_writer()
        dataset = model.examples_inputter.make_evaluation_dataset(
            features_file,
            labels_file,
            batch_size,
            batch_type=batch_type,
            length_bucket_width=length_bucket_width,
            num_threads=1,
            prefetch_buffer_size=1,
        )

        self._eval_fn = tf.function(
            model.evaluate, input_signature=dataset.element_spec
        )
        self._dataset = dataset

        self._metrics_name = {"loss", "perplexity"}
        for scorer in self._scorers:
            self._metrics_name.update(scorer.scores_name)
        model_metrics = self._model.get_metrics()
        if model_metrics:
            self._metrics_name.update(set(model_metrics.keys()))

        if early_stopping is not None:
            if early_stopping.metric not in self._metrics_name:
                raise ValueError(
                    "Invalid early stopping metric '%s', expected one in %s"
                    % (early_stopping.metric, str(self._metrics_name))
                )
            if early_stopping.steps <= 0:
                raise ValueError("Early stopping steps should greater than 0")
        self._early_stopping = early_stopping

        self._export_on_best = export_on_best
        self._exporter = exporter
        self._export_dir = export_dir
        self._max_exports_to_keep = max_exports_to_keep

    @classmethod
    def from_config(cls, model, config, features_file=None, labels_file=None):
        """Creates an evaluator from the configuration.

        Args:
          model: A :class:`opennmt.models.Model` to evaluate.
          config: The global user configuration.
          features_file: Optional input features file to evaluate. If not set, will
            load ``eval_features_file`` from the data configuration.
          labels_file: Optional output labels file to evaluate. If not set, will load
            ``eval_labels_file`` from the data configuration.

        Returns:
          A :class:`opennmt.evaluation.Evaluator` instance.

        Raises:
          ValueError: for supervised models, if one of :obj:`features_file` and
            :obj:`labels_file` is set but not the other.
          ValueError: for unsupervised models, if :obj:`labels_file` is set.
        """
        if model.unsupervised:
            if labels_file is not None:
                raise ValueError(
                    "labels_file can not be set when evaluating unsupervised models"
                )
        elif (features_file is None) != (labels_file is None):
            raise ValueError(
                "features_file and labels_file should be both set for evaluation"
            )
        eval_config = config["eval"]
        scorers = eval_config.get("scorers")
        if scorers is None:
            # Backward compatibility with previous field name.
            scorers = eval_config.get("external_evaluators")
        if scorers is not None:
            scorers = scorers_lib.make_scorers(scorers)
        early_stopping_config = eval_config.get("early_stopping")
        if early_stopping_config is not None:
            early_stopping = EarlyStopping(
                metric=early_stopping_config.get("metric", "loss"),
                min_improvement=early_stopping_config.get("min_improvement", 0),
                steps=early_stopping_config["steps"],
            )
        else:
            early_stopping = None
        return cls(
            model,
            features_file or config["data"]["eval_features_file"],
            labels_file or config["data"].get("eval_labels_file"),
            eval_config["batch_size"],
            batch_type=eval_config.get("batch_type", "examples"),
            length_bucket_width=eval_config.get("length_bucket_width"),
            scorers=scorers,
            save_predictions=eval_config.get("save_eval_predictions", False),
            early_stopping=early_stopping,
            model_dir=config["model_dir"],
            export_on_best=eval_config.get("export_on_best"),
            exporter=exporters.make_exporter(
                eval_config.get("export_format", "saved_model")
            ),
            max_exports_to_keep=eval_config.get("max_exports_to_keep", 5),
        )

    @property
    def predictions_dir(self):
        """The directory containing saved predictions."""
        return self._eval_dir

    @property
    def export_dir(self):
        """The directory containing exported models."""
        return self._export_dir

    @property
    def metrics_name(self):
        """The name of the metrics returned by this evaluator."""
        return self._metrics_name

    @property
    def metrics_history(self):
        """The history of metrics result per evaluation step."""
        return self._metrics_history

    @property
    def last_evaluated_step(self):
        """The last training step that was evaluated."""
        if not self._metrics_history:
            return None
        return self._metrics_history[-1][0]

    def _is_higher_better_for_metric(self, metric):
        # Look if the metric is produced by a scorer as they define the scores order.
        for scorer in self._scorers:
            if metric in scorer.scores_name:
                return scorer.higher_is_better()
        # TODO: the condition below is not always true, find a way to set it
        # correctly for Keras metrics.
        return metric not in ("loss", "perplexity")

    def _get_metric_history(self, metric):
        return [
            metrics[metric] for _, metrics in self._metrics_history if metric in metrics
        ]

    def should_stop(self):
        """Returns ``True`` if early stopping conditions are met."""
        if self._early_stopping is None:
            return False
        target_metric = self._early_stopping.metric
        higher_is_better = self._is_higher_better_for_metric(target_metric)
        metrics = self._get_metric_history(target_metric)
        should_stop = early_stop(
            metrics,
            self._early_stopping.steps,
            min_improvement=self._early_stopping.min_improvement,
            higher_is_better=higher_is_better,
        )
        if should_stop:
            tf.get_logger().warning(
                "Evaluation metric '%s' did not improve more than %f in the last %d evaluations",
                target_metric,
                self._early_stopping.min_improvement,
                self._early_stopping.steps,
            )
        return should_stop

    def is_best(self, metric):
        """Returns ``True`` if the latest value of :obj:`metric` is the best so far.

        Args:
          metric: The metric to consider.

        Returns:
          A boolean.
        """
        metric_history = self._get_metric_history(metric)
        if not metric_history:
            return False
        metric_history, latest_value = metric_history[:-1], metric_history[-1]
        if not metric_history:
            return True
        if self._is_higher_better_for_metric(metric):
            return latest_value > max(metric_history)
        else:
            return latest_value < min(metric_history)

    def __call__(self, step):
        """Runs the evaluator.

        Args:
          step: The current training step.

        Returns:
          A dictionary of evaluation metrics.
        """
        tf.get_logger().info("Running evaluation for step %d", step)
        output_file = None
        output_path = None
        if self._save_predictions:
            output_path = os.path.join(self._eval_dir, "predictions.txt.%d" % step)
            output_file = tf.io.gfile.GFile(output_path, "w")
            params = {"n_best": 1}
            write_fn = lambda prediction: (
                self._model.print_prediction(
                    prediction, params=params, stream=output_file
                )
            )
            index_fn = lambda prediction: prediction.get("index")
            ordered_writer = misc.OrderRestorer(index_fn, write_fn)

        loss_num = 0
        loss_den = 0
        metrics = self._model.get_metrics()
        for source, target in self._dataset:
            loss, predictions = self._eval_fn(source, target)
            if isinstance(loss, tuple):
                loss_num += loss[0]
                loss_den += loss[1]
            else:
                loss_num += loss
                loss_den += 1
            if metrics:
                self._model.update_metrics(metrics, predictions, target)
            if output_file is not None:
                predictions = {k: v.numpy() for k, v in predictions.items()}
                for prediction in misc.extract_batches(predictions):
                    ordered_writer.push(prediction)
        if loss_den == 0:
            raise RuntimeError("No examples were evaluated")
        loss = loss_num / loss_den

        results = dict(loss=loss, perplexity=tf.math.exp(loss))
        if metrics:
            for name, metric in metrics.items():
                results[name] = metric.result()
        if self._save_predictions:
            tf.get_logger().info("Evaluation predictions saved to %s", output_path)
            output_file.close()
            for scorer in self._scorers:
                score = scorer(self._labels_file, output_path)
                if isinstance(score, dict):
                    results.update(score)
                else:
                    results[scorer.name] = score

        for name, value in results.items():
            if isinstance(value, tf.Tensor):
                results[name] = value.numpy()

        self._record_results(step, results)
        self._maybe_export(step, results)
        self._maybe_garbage_collect_exports()
        return results

    def _record_results(self, step, results):
        # Clear history for steps that are greater than step.
        while self._metrics_history and self._metrics_history[-1][0] > step:
            self._metrics_history.pop()
        self._metrics_history.append((step, dict(results)))
        tf.get_logger().info(
            "Evaluation result for step %d: %s",
            step,
            " ; ".join("%s = %f" % (k, v) for k, v in results.items()),
        )
        with self._summary_writer.as_default():
            for key, value in results.items():
                tf.summary.scalar("%s/%s" % (_SUMMARIES_SCOPE, key), value, step=step)
            self._summary_writer.flush()

    def _maybe_export(self, step, results):
        if self._export_on_best is None or not self.is_best(self._export_on_best):
            return
        export_dir = os.path.join(self._export_dir, str(step))
        tf.get_logger().info(
            "Exporting model to %s (best %s so far: %f)",
            export_dir,
            self._export_on_best,
            results[self._export_on_best],
        )
        self._model.export(export_dir, exporter=self._exporter)

    def _maybe_garbage_collect_exports(self):
        if self._max_exports_to_keep is None or not os.path.exists(self._export_dir):
            return
        exported_steps = list(sorted(map(int, os.listdir(self._export_dir))))
        num_exports = len(exported_steps)
        if num_exports > self._max_exports_to_keep:
            steps_to_remove = exported_steps[: num_exports - self._max_exports_to_keep]
            for step in steps_to_remove:
                shutil.rmtree(os.path.join(self._export_dir, str(step)))


def early_stop(metrics, steps, min_improvement=0, higher_is_better=False):
    """Early stopping condition.

    Args:
      metrics: A list of metric values.
      steps: Consider the improvement over this many steps.
      min_improvement: Continue if the metric improved less than this value:
      higher_is_better: Whether a higher value is better for this metric.

    Returns:
      A boolean.
    """
    if len(metrics) < steps + 1:
        return False

    def _did_improve(ref, new):
        # Returns True if new is improving on ref.
        if higher_is_better:
            return new > ref + min_improvement
        else:
            return new < ref - min_improvement

    ref_metric = metrics[-steps - 1]
    for metric in metrics[-steps:]:
        if _did_improve(ref_metric, metric):
            return False
    return True



"""Various tensor manipulation functions."""

import tensorflow as tf


def roll_sequence(tensor, offsets):
    """Shifts sequences by an offset.

    Args:
      tensor: A ``tf.Tensor`` of shape :math:`[B, T, ...]`.
      offsets : The offset of each sequence of shape :math:`[B]`.

    Returns:
      A ``tf.Tensor`` with the same shape as :obj:`tensor` and with sequences
      shifted by :obj:`offsets`.
    """
    batch_size = tf.shape(tensor)[0]
    time = tf.shape(tensor)[1]
    cols, rows = tf.meshgrid(tf.range(time), tf.range(batch_size))
    cols -= tf.expand_dims(offsets, 1)
    cols %= time
    indices = tf.stack([rows, cols], axis=-1)
    return tf.gather_nd(tensor, indices)



"""Module defining reusable and model specific layers."""

from opennmt.layers.bridge import Bridge, CopyBridge, DenseBridge, ZeroBridge
from opennmt.layers.common import Dense, LayerNorm, LayerWrapper, dropout, gelu
from opennmt.layers.position import (
    PositionEmbedder,
    PositionEncoder,
    SinusoidalPositionEncoder,
)
from opennmt.layers.reducer import (
    ConcatReducer,
    DenseReducer,
    JoinReducer,
    MultiplyReducer,
    Reducer,
    SumReducer,
)
from opennmt.layers.rnn import LSTM, RNN, RNNCellWrapper, make_rnn_cell
from opennmt.layers.transformer import (
    FeedForwardNetwork,
    MultiHeadAttention,
    MultiHeadAttentionReduction,
    SelfAttentionDecoderLayer,
    SelfAttentionEncoderLayer,
    TransformerLayerWrapper,
    combine_heads,
    future_mask,
    split_heads,
)



"""Define model exporters."""

import abc
import os
import tempfile

import tensorflow as tf

from opennmt.utils import misc


class Exporter(abc.ABC):
    """Base class for model exporters."""

    def export(self, model, export_dir):
        """Exports :obj:`model` to :obj:`export_dir`.

        Raises:
          ValueError: if :obj:`model` is not supported by this exporter.
        """
        self._export_model(model, export_dir)
        with tempfile.TemporaryDirectory() as tmp_dir:
            extra_assets = model.export_assets(tmp_dir)
            if extra_assets:
                assets_extra = os.path.join(export_dir, "assets.extra")
                tf.io.gfile.makedirs(assets_extra)
                for filename, path in extra_assets.items():
                    tf.io.gfile.copy(
                        path, os.path.join(assets_extra, filename), overwrite=True
                    )
                tf.get_logger().info("Extra assets written to: %s", assets_extra)

    @abc.abstractmethod
    def _export_model(self, model, export_dir):
        raise NotImplementedError()


_EXPORTERS_REGISTRY = misc.ClassRegistry(base_class=Exporter)
register_exporter = _EXPORTERS_REGISTRY.register


def make_exporter(name, **kwargs):
    """Creates a new exporter.

    Args:
      name: The exporter name.
      **kwargs: Additional arguments to pass to the exporter constructor.

    Returns:
      A :class:`opennmt.utils.Exporter` instance.

    Raises:
      ValueError: if :obj:`name` is invalid.
    """
    exporter_class = _EXPORTERS_REGISTRY.get(name)
    if exporter_class is None:
        raise ValueError("Invalid exporter name: %s" % name)
    return exporter_class(**kwargs)


def list_exporters():
    """Lists the name of registered exporters."""
    return _EXPORTERS_REGISTRY.class_names


@register_exporter(name="saved_model")
class SavedModelExporter(Exporter):
    """SavedModel exporter."""

    def _export_model(self, model, export_dir):
        tf.saved_model.save(model, export_dir, signatures=model.serve_function())


@register_exporter(name="checkpoint")
class CheckpointExporter(Exporter):
    """Checkpoint exporter."""

    def _export_model(self, model, export_dir):
        checkpoint = tf.train.Checkpoint(model=model)
        checkpoint.write(os.path.join(export_dir, "ckpt"))


@register_exporter(name="tflite")
class TFLiteExporter(Exporter):
    """TensorFlow Lite exporter."""

    def __init__(self, quantization=None):
        accepted_quantization = (
            "float16",
            "dynamic_range",
        )
        if quantization is not None and quantization not in accepted_quantization:
            raise ValueError(
                "Unsupported quantization '%s' for TensorFlow Lite, accepted values are: %s"
                % (quantization, ", ".join(accepted_quantization))
            )
        self._quantization = quantization

    def _export_model(self, model, export_dir):
        # Tries to run prediction with TensorFlow Lite method it will convert
        tflite_concrete_fn = model.tflite_function().get_concrete_function()

        # Saving
        converter = tf.lite.TFLiteConverter.from_concrete_functions(
            [tflite_concrete_fn]
        )
        converter.target_spec.supported_ops = [
            tf.lite.OpsSet.TFLITE_BUILTINS,  # enable TensorFlow Lite ops.
            tf.lite.OpsSet.SELECT_TF_OPS,  # enable TensorFlow ops.
        ]
        if self._quantization == "dynamic_range":
            converter.optimizations = [tf.lite.Optimize.DEFAULT]
        elif self._quantization == "float16":
            converter.optimizations = [tf.lite.Optimize.DEFAULT]
            converter.target_spec.supported_types = [tf.float16]

        tflite_model_path = os.path.join(export_dir, "opennmt.tflite")
        tflite_model = converter.convert()
        with tf.io.gfile.GFile(tflite_model_path, "wb") as f:
            f.write(tflite_model)


@register_exporter(name="tflite_float16")
class TFLiteFloat16Exporter(TFLiteExporter):
    """TensorFlow Lite exporter with float16 quantization."""

    def __init__(self):
        super().__init__(quantization="float16")


@register_exporter(name="tflite_dynamic_range")
class TFLiteDynamicRangeExporter(TFLiteExporter):
    """TensorFlow Lite exporter with dynamic range quantization."""

    def __init__(self):
        super().__init__(quantization="dynamic_range")


@register_exporter(name="ctranslate2")
class CTranslate2Exporter(Exporter):
    """CTranslate2 exporter."""

    def __init__(self, quantization=None):
        """Initializes the exporter.

        Args:
          quantization: Quantize model weights to this type when exporting the model.
            Can be "int8", "int16", or "float16". Default is no quantization.

        Raises:
          ImportError: if the CTranslate2 package is missing.
          ValueError: if :obj:`quantization` is invalid.
        """
        # Fail now if ctranslate2 package is missing.
        import ctranslate2

        accepted_quantization = ("int8", "int16", "float16", "int8_float16")
        if quantization is not None and quantization not in accepted_quantization:
            raise ValueError(
                "Invalid quantization '%s' for CTranslate2, accepted values are: %s"
                % (quantization, ", ".join(accepted_quantization))
            )
        self._quantization = quantization

    def _export_model(self, model, export_dir):
        model_spec = model.ctranslate2_spec
        if model_spec is None:
            raise ValueError(
                "The model does not define an equivalent CTranslate2 model specification"
            )
        if not model.built:
            model.create_variables()
        variables = misc.get_variables_name_mapping(model, root_key="model")
        variables = {
            name.replace("/.ATTRIBUTES/VARIABLE_VALUE", ""): value.numpy()
            for name, value in variables.items()
        }
        import ctranslate2

        converter = ctranslate2.converters.OpenNMTTFConverter(
            model_spec,
            model.features_inputter.vocabulary_file,
            model.labels_inputter.vocabulary_file,
            variables=variables,
        )
        converter.convert(export_dir, quantization=self._quantization, force=True)


@register_exporter(name="ctranslate2_int8")
class CTranslate2Int8Exporter(CTranslate2Exporter):
    """CTranslate2 exporter with int8 quantization."""

    def __init__(self):
        super().__init__(quantization="int8")


@register_exporter(name="ctranslate2_int16")
class CTranslate2Int16Exporter(CTranslate2Exporter):
    """CTranslate2 exporter with int16 quantization."""

    def __init__(self):
        super().__init__(quantization="int16")


@register_exporter(name="ctranslate2_float16")
class CTranslate2Float16Exporter(CTranslate2Exporter):
    """CTranslate2 exporter with float16 quantization."""

    def __init__(self):
        super().__init__(quantization="float16")


@register_exporter(name="ctranslate2_int8_float16")
class CTranslate2Int8Float16Exporter(CTranslate2Exporter):
    """CTranslate2 exporter with int8_float16 quantization."""

    def __init__(self):
        super().__init__(quantization="int8_float16")



"""Define the OpenNMT tokenizer."""

import os

import pyonmttok
import tensorflow as tf
import yaml

from opennmt.tokenizers import tokenizer


@tokenizer.register_tokenizer
class OpenNMTTokenizer(tokenizer.Tokenizer):
    """Tokenizer based on the OpenNMT Tokenizer: https://github.com/OpenNMT/Tokenizer."""

    def __init__(self, **kwargs):
        """Initializes the tokenizer.

        Args:
          **kwargs: Tokenization options, see
            https://github.com/OpenNMT/Tokenizer/blob/master/docs/options.md.
        """
        case_feature = kwargs.get("case_feature")
        if case_feature:
            raise ValueError("case_feature is not supported with OpenNMT-tf")
        kwargs.setdefault("mode", "conservative")
        self._config = kwargs
        self._tokenizer = pyonmttok.Tokenizer(**kwargs)

    @property
    def config(self):
        """The tokenization configuration."""
        return self._config.copy()

    @property
    def opennmt_tokenizer(self):
        """The ``pyonmttok.Tokenizer`` instance."""
        return self._tokenizer

    def export_assets(self, asset_dir, asset_prefix=""):
        assets = {}

        # Extract asset files from the configuration.
        config = self._config.copy()
        for key, value in config.items():
            if isinstance(value, str) and tf.io.gfile.exists(value):
                basename = os.path.basename(value)
                config[key] = basename  # Only save the basename.
                assets[basename] = value

        # Save the tokenizer configuration.
        config_name = "%stokenizer_config.yml" % asset_prefix
        config_path = os.path.join(asset_dir, config_name)
        assets[config_name] = config_path
        with tf.io.gfile.GFile(config_path, "w") as config_file:
            yaml.dump(config, stream=config_file, default_flow_style=False)

        return assets

    def _tokenize_string(self, text, training):
        tokens, _ = self._tokenizer.tokenize(text, training=training)
        return tokens

    def _tokenize_string_batch(self, batch_text, training):
        tokens, _ = self._tokenizer.tokenize_batch(batch_text, training=training)
        return tokens

    def _detokenize_string(self, tokens):
        return self._tokenizer.detokenize(tokens)



"""Main script."""

import argparse
import logging
import os
import sys

import tensorflow as tf

from opennmt import __version__
from opennmt import config as config_util
from opennmt.models import catalog
from opennmt.runner import Runner
from opennmt.utils import exporters


def _initialize_logging(log_level):
    logger = tf.get_logger()
    logger.setLevel(log_level)

    # Configure the TensorFlow logger to use the same log format as the TensorFlow C++ logs.
    for handler in list(logger.handlers):
        logger.removeHandler(handler)
    formatter = logging.Formatter(
        fmt="%(asctime)s.%(msecs)03d000: %(levelname).1s %(filename)s:%(lineno)d] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-v", "--version", action="version", version="OpenNMT-tf %s" % __version__
    )
    parser.add_argument(
        "--config", required=True, nargs="+", help="List of configuration files."
    )
    parser.add_argument(
        "--model_dir",
        help=(
            "Path to the model directory. If not set, the model directory is read "
            "from the field 'model_dir' in the configuration."
        ),
    )
    parser.add_argument(
        "--auto_config",
        default=None,
        action="store_true",
        help="Enable automatic configuration values.",
    )
    parser.add_argument(
        "--model_type",
        default="",
        choices=list(sorted(catalog.list_model_names_from_catalog())),
        help="Model type from the catalog.",
    )
    parser.add_argument("--model", default="", help="Custom model configuration file.")
    parser.add_argument(
        "--run_dir",
        default="",
        help="If set, model_dir will be created relative to this location.",
    )
    parser.add_argument(
        "--data_dir",
        default="",
        help="If set, data files are expected to be relative to this location.",
    )
    parser.add_argument(
        "--checkpoint_path",
        default=None,
        help=(
            "Path to the checkpoint or checkpoint directory to load. "
            "If not set, the latest checkpoint from the model directory is loaded."
        ),
    )
    parser.add_argument(
        "--log_level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
        help="Logs verbosity.",
    )
    parser.add_argument("--seed", type=int, default=None, help="Random seed.")
    parser.add_argument(
        "--gpu_allow_growth",
        default=False,
        action="store_true",
        help="Allocate GPU memory dynamically.",
    )
    parser.add_argument(
        "--intra_op_parallelism_threads",
        type=int,
        default=0,
        help=(
            "Number of intra op threads (0 means the system picks "
            "an appropriate number)."
        ),
    )
    parser.add_argument(
        "--inter_op_parallelism_threads",
        type=int,
        default=0,
        help=(
            "Number of inter op threads (0 means the system picks "
            "an appropriate number)."
        ),
    )
    parser.add_argument(
        "--mixed_precision",
        default=False,
        action="store_true",
        help="Enable mixed precision.",
    )
    parser.add_argument(
        "--eager_execution",
        default=False,
        action="store_true",
        help="Enable TensorFlow eager execution.",
    )

    subparsers = parser.add_subparsers(help="Run type.", dest="run_type")
    subparsers.required = True
    parser_train = subparsers.add_parser("train", help="Training.")
    parser_train.add_argument(
        "--with_eval",
        default=False,
        action="store_true",
        help="Enable automatic evaluation.",
    )
    parser_train.add_argument(
        "--num_gpus",
        type=int,
        default=1,
        help="Number of GPUs to use for in-graph replication.",
    )
    parser_train.add_argument(
        "--horovod",
        default=False,
        action="store_true",
        help="Enable Horovod training mode.",
    )
    parser_train.add_argument(
        "--continue_from_checkpoint",
        default=False,
        action="store_true",
        help=(
            "Continue the training from the checkpoint passed to --checkpoint_path. "
            "If --checkpoint_path is set but not --continue_from_checkpoint, only the "
            "model weights are loaded and the optimization states are ignored."
        ),
    )

    parser_eval = subparsers.add_parser("eval", help="Evaluation.")
    parser_eval.add_argument(
        "--features_file", nargs="+", default=None, help="Input features files."
    )
    parser_eval.add_argument("--labels_file", default=None, help="Output labels files.")

    parser_infer = subparsers.add_parser("infer", help="Inference.")
    parser_infer.add_argument(
        "--features_file", nargs="+", required=True, help="Run inference on this file."
    )
    parser_infer.add_argument(
        "--predictions_file",
        default="",
        help=(
            "File used to save predictions. If not set, predictions are printed "
            "on the standard output."
        ),
    )
    parser_infer.add_argument(
        "--log_prediction_time",
        default=False,
        action="store_true",
        help="Logs some prediction time metrics.",
    )

    parser_export = subparsers.add_parser("export", help="Model export.")
    parser_export.add_argument(
        "--output_dir",
        "--export_dir",
        required=True,
        help="The directory of the exported model.",
    )
    parser_export.add_argument(
        "--format",
        "--export_format",
        choices=exporters.list_exporters(),
        default="saved_model",
        help="Format of the exported model.",
    )

    parser_score = subparsers.add_parser("score", help="Scoring.")
    parser_score.add_argument(
        "--features_file", nargs="+", required=True, help="Features file."
    )
    parser_score.add_argument(
        "--predictions_file", default=None, help="Predictions to score."
    )
    parser_score.add_argument("--output_file", default=None, help="Output file.")

    parser_average_checkpoints = subparsers.add_parser(
        "average_checkpoints", help="Checkpoint averaging."
    )
    parser_average_checkpoints.add_argument(
        "--output_dir",
        required=True,
        help="The output directory for the averaged checkpoint.",
    )
    parser_average_checkpoints.add_argument(
        "--max_count",
        type=int,
        default=8,
        help="The maximal number of checkpoints to average.",
    )

    parser_update_vocab = subparsers.add_parser(
        "update_vocab", help="Update model vocabularies in checkpoint."
    )
    parser_update_vocab.add_argument(
        "--output_dir",
        required=True,
        help="The output directory for the updated checkpoint.",
    )
    parser_update_vocab.add_argument(
        "--src_vocab", default=None, help="Path to the new source vocabulary."
    )
    parser_update_vocab.add_argument(
        "--tgt_vocab", default=None, help="Path to the new target vocabulary."
    )

    # When using an option that takes multiple values just before the run type,
    # the run type is treated as a value of this option. To fix this issue, we
    # inject a placeholder option just before the run type to clearly separate it.
    parser.add_argument("--placeholder", action="store_true", help=argparse.SUPPRESS)
    run_types = set(subparsers.choices.keys())
    args = sys.argv[1:]
    for i, arg in enumerate(args):
        if arg in run_types:
            args.insert(i, "--placeholder")
            break

    args = parser.parse_args(args)
    if (
        hasattr(args, "features_file")
        and args.features_file
        and len(args.features_file) == 1
    ):
        args.features_file = args.features_file[0]

    _initialize_logging(getattr(logging, args.log_level))
    tf.config.threading.set_intra_op_parallelism_threads(
        args.intra_op_parallelism_threads
    )
    tf.config.threading.set_inter_op_parallelism_threads(
        args.inter_op_parallelism_threads
    )

    if args.eager_execution:
        tf.config.run_functions_eagerly(True)

    gpus = tf.config.list_physical_devices(device_type="GPU")
    if hasattr(args, "horovod") and args.horovod:
        import horovod.tensorflow as hvd

        hvd.init()
        is_master = hvd.rank() == 0
        if gpus:
            local_gpu = gpus[hvd.local_rank()]
            tf.config.set_visible_devices(local_gpu, device_type="GPU")
            gpus = [local_gpu]
    else:
        hvd = None
        is_master = True

    if args.gpu_allow_growth:
        for device in gpus:
            tf.config.experimental.set_memory_growth(device, enable=True)

    # Load and merge run configurations.
    config = config_util.load_config(args.config)
    if args.model_dir:
        config["model_dir"] = args.model_dir
    elif "model_dir" not in config:
        raise ValueError(
            "No model directory is defined: you should either set --model_dir "
            "on the command line or set the field model_dir in the configuration"
        )
    if args.run_dir:
        config["model_dir"] = os.path.join(args.run_dir, config["model_dir"])
    if args.data_dir:
        config["data"] = config_util.try_prefix_paths(args.data_dir, config["data"])

    if is_master and not tf.io.gfile.exists(config["model_dir"]):
        tf.get_logger().info("Creating model directory %s", config["model_dir"])
        tf.io.gfile.makedirs(config["model_dir"])

    model = config_util.load_model(
        config["model_dir"],
        model_file=args.model,
        model_name=args.model_type,
        serialize_model=is_master,
        as_builder=True,
    )
    runner = Runner(
        model,
        config,
        auto_config=args.auto_config,
        mixed_precision=args.mixed_precision,
        seed=args.seed,
    )

    if args.run_type == "train":
        runner.train(
            num_devices=args.num_gpus,
            with_eval=args.with_eval,
            checkpoint_path=args.checkpoint_path,
            hvd=hvd,
            continue_from_checkpoint=args.continue_from_checkpoint,
        )
    elif args.run_type == "eval":
        metrics = runner.evaluate(
            checkpoint_path=args.checkpoint_path,
            features_file=args.features_file,
            labels_file=args.labels_file,
        )
        print(metrics)
    elif args.run_type == "infer":
        runner.infer(
            args.features_file,
            predictions_file=args.predictions_file,
            checkpoint_path=args.checkpoint_path,
            log_time=args.log_prediction_time,
        )
    elif args.run_type == "export":
        runner.export(
            args.output_dir,
            checkpoint_path=args.checkpoint_path,
            exporter=exporters.make_exporter(args.format),
        )
    elif args.run_type == "score":
        runner.score(
            args.features_file,
            args.predictions_file,
            checkpoint_path=args.checkpoint_path,
            output_file=args.output_file,
        )
    elif args.run_type == "average_checkpoints":
        runner.average_checkpoints(args.output_dir, max_count=args.max_count)
    elif args.run_type == "update_vocab":
        runner.update_vocab(
            args.output_dir, src_vocab=args.src_vocab, tgt_vocab=args.tgt_vocab
        )


if __name__ == "__main__":
    main()



"""Define bridges: logic of passing the encoder state to the decoder."""

import abc

import tensorflow as tf


def assert_state_is_compatible(expected_state, state):
    """Asserts that states are compatible.

    Args:
      expected_state: The reference state.
      state: The state that must be compatible with :obj:`expected_state`.

    Raises:
      ValueError: if the states are incompatible.
    """
    # Check structure compatibility.
    tf.nest.assert_same_structure(expected_state, state)

    # Check shape compatibility.
    expected_state_flat = tf.nest.flatten(expected_state)
    state_flat = tf.nest.flatten(state)

    for x, y in zip(expected_state_flat, state_flat):
        if tf.is_tensor(x):
            expected_depth = x.shape[-1]
            depth = y.shape[-1]
            if depth != expected_depth:
                raise ValueError(
                    "Tensor in state has shape %s which is incompatible "
                    "with the target shape %s" % (y.shape, x.shape)
                )


class Bridge(tf.keras.layers.Layer):
    """Base class for bridges."""

    def __call__(self, encoder_state, decoder_zero_state):
        """Returns the initial decoder state.

        Args:
          encoder_state: The encoder state.
          decoder_zero_state: The default decoder state.

        Returns:
          The decoder initial state.
        """
        return super().__call__([encoder_state, decoder_zero_state])

    @abc.abstractmethod
    def call(self, states):
        raise NotImplementedError()


class CopyBridge(Bridge):
    """A bridge that passes the encoder state as is."""

    def call(self, states):
        encoder_state, decoder_state = states
        assert_state_is_compatible(encoder_state, decoder_state)
        flat_encoder_state = tf.nest.flatten(encoder_state)
        return tf.nest.pack_sequence_as(decoder_state, flat_encoder_state)


class ZeroBridge(Bridge):
    """A bridge that does not pass information from the encoder."""

    def call(self, states):
        # Simply return the default decoder state.
        return states[1]


class DenseBridge(Bridge):
    """A bridge that applies a parameterized linear transformation from the
    encoder state to the decoder state size.
    """

    def __init__(self, activation=None):
        """Initializes the bridge.

        Args:
          activation: Activation function (a callable).
            Set it to ``None`` to maintain a linear activation.
        """
        super().__init__()
        self.activation = activation
        self.decoder_state_sizes = None
        self.linear = None

    def build(self, input_shape):
        decoder_shape = input_shape[1]
        self.decoder_state_sizes = [
            shape[-1] for shape in tf.nest.flatten(decoder_shape)
        ]
        self.linear = tf.keras.layers.Dense(
            sum(self.decoder_state_sizes), activation=self.activation
        )

    def call(self, states):
        encoder_state, decoder_state = states
        encoder_state_flat = tf.nest.flatten(encoder_state)
        encoder_state_single = tf.concat(encoder_state_flat, 1)
        transformed = self.linear(encoder_state_single)
        splitted = tf.split(transformed, self.decoder_state_sizes, axis=1)
        return tf.nest.pack_sequence_as(decoder_state, splitted)



"""Define RNN-based encoders."""

import tensorflow as tf
import tensorflow_addons as tfa

from opennmt.encoders.encoder import Encoder, SequentialEncoder
from opennmt.layers import common, rnn
from opennmt.layers.reducer import ConcatReducer, JoinReducer, pad_in_time


class _RNNEncoderBase(Encoder):
    """Base class for RNN encoders."""

    def __init__(self, rnn_layer, **kwargs):
        """Initializes the encoder.

        Args:
          rnn_layer: The RNN layer used to encode the inputs.
          **kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        self.rnn = rnn_layer

    def call(self, inputs, sequence_length=None, training=None):
        mask = self.build_mask(inputs, sequence_length=sequence_length)
        outputs, states = self.rnn(inputs, mask=mask, training=training)
        return outputs, states, sequence_length


class RNNEncoder(_RNNEncoderBase):
    """A RNN sequence encoder."""

    def __init__(
        self,
        num_layers,
        num_units,
        bidirectional=False,
        residual_connections=False,
        dropout=0.3,
        reducer=ConcatReducer(),
        cell_class=None,
        **kwargs
    ):
        """Initializes the parameters of the encoder.

        Args:
          num_layers: The number of layers.
          num_units: The number of units in each layer.
          bidirectional: Use a bidirectional RNN.
          residual_connections: If ``True``, each layer input will be added to its
            output.
          dropout: The probability to drop units in each layer output.
          reducer: A :class:`opennmt.layers.Reducer` instance to merge
            bidirectional state and outputs.
          cell_class: The inner cell class or a callable taking :obj:`num_units` as
            argument and returning a cell. Defaults to a LSTM cell.
          **kwargs: Additional layer arguments.
        """
        cell = rnn.make_rnn_cell(
            num_layers,
            num_units,
            dropout=dropout,
            residual_connections=residual_connections,
            cell_class=cell_class,
        )
        rnn_layer = rnn.RNN(cell, bidirectional=bidirectional, reducer=reducer)
        super().__init__(rnn_layer, **kwargs)

    def map_v1_weights(self, weights):
        return self.rnn.map_v1_weights(weights)


class LSTMEncoder(_RNNEncoderBase):
    """A LSTM sequence encoder.

    See Also:
      :class:`opennmt.layers.LSTM` for differences between this encoder and
      :class:`opennmt.encoders.RNNEncoder` with a `LSTMCell`.
    """

    def __init__(
        self,
        num_layers,
        num_units,
        bidirectional=False,
        residual_connections=False,
        dropout=0.3,
        reducer=ConcatReducer(),
        **kwargs
    ):
        """Initializes the parameters of the encoder.

        Args:
          num_layers: The number of layers.
          num_units: The number of units in each layer output.
          bidirectional: Make each LSTM layer bidirectional.
          residual_connections: If ``True``, each layer input will be added to its
            output.
          dropout: The probability to drop units in each layer output.
          reducer: A :class:`opennmt.layers.Reducer` instance to merge
            bidirectional state and outputs.
          **kwargs: Additional layer arguments.
        """
        lstm_layer = rnn.LSTM(
            num_layers,
            num_units,
            bidirectional=bidirectional,
            reducer=reducer,
            dropout=dropout,
            residual_connections=residual_connections,
        )
        super().__init__(lstm_layer, **kwargs)


class GNMTEncoder(SequentialEncoder):
    """The RNN encoder used in GNMT as described in
    https://arxiv.org/abs/1609.08144.
    """

    def __init__(self, num_layers, num_units, dropout=0.3):
        """Initializes the parameters of the encoder.

        Args:
          num_layers: The number of layers.
          num_units: The number of units in each layer.
          dropout: The probability to drop units in each layer output.

        Raises:
          ValueError: if :obj:`num_layers` < 2.
        """
        if num_layers < 2:
            raise ValueError("GNMTEncoder requires at least 2 layers")
        bidirectional = LSTMEncoder(1, num_units, bidirectional=True, dropout=dropout)
        unidirectional = LSTMEncoder(
            num_layers - 1, num_units, dropout=dropout, residual_connections=True
        )
        super().__init__([bidirectional, unidirectional])


class RNMTPlusEncoder(SequentialEncoder):
    """The RNMT+ encoder described in https://arxiv.org/abs/1804.09849."""

    def __init__(self, num_layers=6, num_units=1024, cell_class=None, dropout=0.3):
        """Initializes the parameters of the encoder.

        Args:
          num_layers: The number of layers.
          num_units: The number of units in each RNN layer and the final output.
          cell_class: The inner cell class or a callable taking :obj:`num_units` as
            argument and returning a cell. Defaults to a layer normalized LSTM cell.
          dropout: The probability to drop units in each layer output.
        """
        if cell_class is None:
            cell_class = tfa.rnn.LayerNormLSTMCell
        layers = [
            RNNEncoder(
                1, num_units, bidirectional=True, dropout=0.0, cell_class=cell_class
            )
            for _ in range(num_layers)
        ]
        layers = [
            common.LayerWrapper(layer, output_dropout=dropout, residual_connection=True)
            for layer in layers
        ]
        super().__init__(layers)
        self.dropout = dropout
        self.projection = tf.keras.layers.Dense(num_units)

    def call(self, inputs, sequence_length=None, training=None):
        inputs = common.dropout(inputs, self.dropout, training=training)
        outputs, state, sequence_length = super().call(
            inputs, sequence_length=sequence_length, training=training
        )
        projected = self.projection(outputs)
        return (projected, state, sequence_length)


class PyramidalRNNEncoder(Encoder):
    """An encoder that reduces the time dimension after each bidirectional layer."""

    def __init__(
        self, num_layers, num_units, reduction_factor=2, cell_class=None, dropout=0.3
    ):
        """Initializes the parameters of the encoder.

        Args:
          num_layers: The number of layers.
          num_units: The number of units in each layer.
          reduction_factor: The time reduction factor.
          cell_class: The inner cell class or a callable taking :obj:`num_units` as
            argument and returning a cell. Defaults to a LSTM cell.
          dropout: The probability to drop units in each layer output.
        """
        super().__init__()
        self.reduction_factor = reduction_factor
        self.state_reducer = JoinReducer()
        self.layers = [
            RNNEncoder(
                1,
                num_units // 2,
                bidirectional=True,
                reducer=ConcatReducer(),
                cell_class=cell_class,
                dropout=dropout,
            )
            for _ in range(num_layers)
        ]

    def call(self, inputs, sequence_length=None, training=None):
        encoder_state = []

        for layer_index, layer in enumerate(self.layers):
            input_depth = inputs.shape[-1]

            if layer_index == 0:
                # For the first input, make the number of timesteps a multiple of the
                # total reduction factor.
                total_reduction_factor = pow(
                    self.reduction_factor, len(self.layers) - 1
                )

                current_length = tf.shape(inputs)[1]
                factor = tf.cast(current_length, tf.float32) / total_reduction_factor
                new_length = (
                    tf.cast(tf.math.ceil(factor), tf.int32) * total_reduction_factor
                )
                inputs = pad_in_time(inputs, new_length - current_length)

                # Lengths should not be smaller than the total reduction factor.
                sequence_length = tf.maximum(sequence_length, total_reduction_factor)
            else:
                # In other cases, reduce the time dimension.
                inputs = tf.reshape(
                    inputs,
                    [tf.shape(inputs)[0], -1, input_depth * self.reduction_factor],
                )
                if sequence_length is not None:
                    sequence_length //= self.reduction_factor

            outputs, state, sequence_length = layer(
                inputs, sequence_length=sequence_length, training=training
            )

            encoder_state.append(state)
            inputs = outputs

        return (outputs, self.state_reducer(encoder_state), sequence_length)



"""Module defining encoders."""

from opennmt.encoders.conv_encoder import ConvEncoder
from opennmt.encoders.encoder import Encoder, ParallelEncoder, SequentialEncoder
from opennmt.encoders.mean_encoder import MeanEncoder
from opennmt.encoders.rnn_encoder import (
    GNMTEncoder,
    LSTMEncoder,
    PyramidalRNNEncoder,
    RNMTPlusEncoder,
    RNNEncoder,
)
from opennmt.encoders.self_attention_encoder import SelfAttentionEncoder



"""Hypotheses file scoring."""


def wer(ref_path, hyp_path):
    """Compute Word Error Rate between two files"""
    with open(ref_path) as ref_fp, open(hyp_path) as hyp_fp:
        wer_score = 0.0
        line_cpt = 0.0
        for ref_line, hyp_line in zip(ref_fp, hyp_fp):
            wer_score += sentence_wer(
                ref_line.strip().split(), hyp_line.strip().split()
            )
            line_cpt = line_cpt + 1
    mean_wer = 1.0
    if line_cpt > 0:
        mean_wer = wer_score / line_cpt
    return mean_wer


def sentence_wer(ref_sent, hyp_sent):
    """Compute Word Error Rate between two sentences (as list of words)"""
    rows = len(ref_sent) + 1
    cols = len(hyp_sent) + 1
    mwer = [[0 for _ in range(cols)] for _ in range(rows)]
    for i in range(rows):
        mwer[i][0] = i
    for j in range(cols):
        mwer[0][j] = j
    for i in range(1, rows):
        for j in range(1, cols):
            if ref_sent[i - 1] == hyp_sent[j - 1]:
                mwer[i][j] = mwer[i - 1][j - 1]
            else:
                substitute = mwer[i - 1][j - 1] + 1
                insert = mwer[i][j - 1] + 1
                delete = mwer[i - 1][j] + 1
                mwer[i][j] = min(substitute, insert, delete)
    sent_wer = 1.0
    if len(ref_sent) > 0:
        sent_wer = mwer[len(ref_sent)][len(hyp_sent)] / len(ref_sent)
    return sent_wer
